/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.extractor.mysql;

import java.sql.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.xquark.extractor.common.MetaDataException;
import org.xquark.extractor.metadata.Catalog;
import org.xquark.extractor.metadata.QNameEncoder;
import org.xquark.extractor.runtime.Selection;
import org.xquark.jdbc.typing.DBMSInfo;
import org.xquark.jdbc.typing.JDBCProperties;

public final class Loader extends org.xquark.extractor.metadata.Loader {

	private static final String RCSRevision = "$Revision: 1.7 $";
	private static final String RCSName     = "$Name:  $";

	public Loader(Connection connection, DBMSInfo dbmsInfo, QNameEncoder encoder)
	throws SQLException {
		super(connection, dbmsInfo, encoder);
	}

	protected String getDefaultSchema() throws SQLException {
		return null;
	}
	
	/**
	 * Create the Catalog metadata node for a specific catalog. Browse JDBC
	 * schemas calling sub-procedure for schemas if matching the selection.
	 * @param catalogName The catalog browsed.
	 * @param selection A user-defined filter used for schemas.
	 * @return the {@link org.xquark.extractor.metadata.Catalog Catalog} node
	 * built.
	 * @throws MetaDataException
	 */
	protected Catalog loadCatalog(String catalogName, Selection selection, JDBCProperties JDBCprop) throws MetaDataException {
		//Trace.enter(this, "loadCatalog(String name, Selection selection)", catalogName);

		List children = new ArrayList();
		List subSelections = new ArrayList();
		ResultSet resultSet = null;
		String name = null;

		/* Catalog node creation */
		Catalog retVal = null;
		if (catalogName == null || catalogName.length() == 0)
			retVal = new Catalog(null);
		else
			retVal = new Catalog(catalogName);

		/* JDBC schema browsing */
		try {
			String defSchema = getDefaultSchema();

			if (selection.noSubElementDef()) {
				Selection schemaSel = new Selection(defSchema, null, Selection.INCLUDES, new HashMap());
				selection.addToMap(defSchema, schemaSel);
			}

			resultSet = _databaseMetaData.getCatalogs();
			Selection subSelection = null;
			while (resultSet.next()) {
				name = resultSet.getString("TABLE_CAT");
				subSelection = selection.includesSubElement(name);
				if (subSelection != null) {
					String targetNameSpace = subSelection.getTargetNamespace();
					//                        if (targetNameSpace == null) targetNameSpace = "";
					if (_nameSpaces.contains(targetNameSpace))
						throw new MetaDataException("Same target namespace : [" + targetNameSpace + "] used twice");
					_nameSpaces.add(targetNameSpace);
					children.add(name);
					subSelections.add(subSelection);
				}
			}
			resultSet.close();

			/* Case where the catalog name is unspecified */
			subSelection = selection.includesSubElement(null);
			if (subSelection != null) {
				if (defSchema == null) {			
					children.add(null);
					subSelections.add(subSelection);
				} else {
					children.add(defSchema);
					subSelection.setNameForDefault(defSchema);
					subSelections.add(subSelection);
				}
			}

			if (children.isEmpty()) {
				throw new MetaDataException("Wrong schema configuration specification. Check your configuration file.");
			}

			/* Browse filtered schema for building */
			for (int i = 0; i < children.size(); i++) {
				String child = (String) children.get(i);
				subSelection = (Selection) subSelections.get(i);
				retVal.addChild(loadSchema(catalogName, child, subSelection, JDBCprop));
			}
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}

		//Trace.exit(this, "loadCatalog(String name, Selection selection)", catalogName);
		return retVal;

	}

}
