/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2004 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;

import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.typing.TypeException;
import org.xquark.xquery.typing.TypeVisitor;

public class JavaModule extends XQueryModule {

    private String className = null;

    private Class javaClass = null;

    private ArrayList javaClassMethods = new ArrayList();

    public JavaModule(String prefix, String namespace, TypeVisitor typeVisitor) throws XQueryException {
        this.staticContext = typeVisitor.getStaticContext();
        this.className = namespace.substring(5);
        // create all stuff necessary
        // create function declaration
        try {
            javaClass = Class.forName(className);
            Method methods[] = javaClass.getMethods();
            ArrayList funcDeclList = new ArrayList();
            HashMap funcDeclMap = new HashMap();
            for (int i = 0; i < methods.length; i++) {
                Method method = (Method) methods[i];
                if (method.getDeclaringClass() == Object.class)
                    continue;
                javaClassMethods.add(method);
                String name = method.getName();
                Class parameterTypes[] = method.getParameterTypes();
                ArrayList parameters = null;
                if (parameterTypes.length > 0) {
                    parameters = new ArrayList(parameterTypes.length);
                    for (int j = 0; j < parameterTypes.length; j++) {
                        Class parameterType = (Class) parameterTypes[j];
                        QName paramName = new QName("param" + (j + 1), null);
                        Variable varj = new Variable(paramName, null, null);
                        SequenceType seqType = getSequenceTypeFromJavaType(parameterType);
                        varj.setTypeDeclaration(seqType);
                        varj.setBindingType(Constants.PARAMETER_BINDINGTYPE);
                        varj.accept(typeVisitor);
                        parameters.add(varj);
                    }
                }
                Class returnType = method.getReturnType();
                SequenceType seqType = getSequenceTypeFromJavaType(returnType);

                FunctionDeclaration funcDecl = new FunctionDeclaration(new QName(namespace, prefix, name, null, null), parameters, seqType, null, this);
                funcDecl.setMethod(method);
                funcDeclList.add(funcDecl);
                funcDeclMap.put(funcDecl.getFuncName().getLocalName(), funcDecl);
            }
            HashMap map = new HashMap();
            map.put("java:" + className, funcDeclMap);
            this.setFunctions(map);
            this.setDeclList(funcDeclList);
        } catch (ClassNotFoundException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    public Class getJavaClass() {
        return javaClass;
    }

    private final String XSPREFIX = "xs";

    private final String XSURI = "http://www.w3.org/2001/XMLSchema";

    /*
     * private QName getQNameFromJavaType(Class javaType) throws TypeException, XQueryException { if (javaType.getName().equals("java.lang.String")) return new QName(XSURI, XSPREFIX, "string", null, null); else if (javaType.getName().equals("int")) return new QName(XSURI, XSPREFIX, "integer", null, null); else if (javaType.getName().equals("org.xquark.xpath.datamodel.TypedValue")) return new QName(XSURI, XSPREFIX, "integer", null, null); else if (javaType.getName().equals("org.xquark.xpath.datamodel.TypedNode")) return new QName(XSURI, XSPREFIX, "integer", null, null); return null; }
     */
    private SequenceType getSequenceTypeFromJavaType(Class javaType) throws TypeException, XQueryException {
        XQueryExpression retExpr = null;
        if (javaType.getName().equals("java.lang.String"))
            retExpr = new QName(XSURI, XSPREFIX, "string", null, null);
        else if (javaType.getName().equals("int"))
            retExpr = new QName(XSURI, XSPREFIX, "integer", null, null);
        else if (javaType.getName().equals("double"))
            retExpr = new QName(XSURI, XSPREFIX, "double", null, null);
        else if (javaType.getName().equals("float"))
            retExpr = new QName(XSURI, XSPREFIX, "float", null, null);
        else if (javaType.getName().equals("boolean"))
            retExpr = new QName(XSURI, XSPREFIX, "boolean", null, null);
        /*
         * else if (javaType.getName().equals("org.xquark.xpath.datamodel.TypedValue")) retExpr = new QName(XSURI, XSPREFIX, "integer", null, null); else if (javaType.getName().equals("org.xquark.xpath.datamodel.TypedNode")) retExpr = new QName(XSURI, XSPREFIX, "integer", null, null);
         */
        if (retExpr == null)
            return null;
        ItemType itemType = new ItemType(retExpr, this);
        SequenceType seqType = new SequenceType(itemType, Constants.NOTHING, this);
        return seqType;
    }

}
