/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.normalize;

import org.xquark.xpath.NodeKind;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionDATA;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionEXISTS;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.FunctionSTRING;

// Note :
// the visitor return value : resultExpr is null if the expression
// has been transformed without creating a new expression

public class ExpressionComparatorVisitor extends DefaultParserVisitor {

    private static final String RCSRevision = "$Revision: 1.3 $";

    private static final String RCSName = "$Name:  $";

    private XQueryExpression expression = null;

    private boolean acceptData = true;

    //private boolean acceptText = true;

    private boolean acceptExists = true;

    private boolean boolResult = false;

    public void setMainExpression(XQueryExpression expr) {
        this.expression = expr;
        if (acceptData && expression instanceof FunctionDATA)
            expression = ((FunctionDATA) expression).getArgument(0);
        if (acceptExists && expression instanceof FunctionEXISTS)
            expression = ((FunctionEXISTS) expression).getArgument(0);
    }

    public XQueryExpression getMainExpression() {
        return expression;
    }

    public boolean getResult() {
        return boolResult;
    }

    // compare expression and value of attribute
    public void visit(AttributeValuePair arg) throws XQueryException {
        boolResult = false;
        arg.getAttributeValue().accept(this);
    }

    public void visit(LocatedExpression arg) throws XQueryException {
        boolResult = false;
        if (expression instanceof LocatedExpression) {
            LocatedExpression loc = (LocatedExpression) expression;
            int sizeLoc = loc.getSteps().size();
            int sizeArg = arg.getSteps().size();
            int i, j;
            for (i = 0, j = 0; i < sizeLoc && j < sizeArg; i++, j++) {
                expression = loc.getStepNum(i);
                arg.getStepNum(j).accept(this);
                if (!boolResult)
                    break;
            }
            if (i == sizeLoc) {
                if (j == sizeArg)
                    boolResult = true;
                else if (j == sizeArg - 1) {
                    Step lastArgStep = arg.getStepNum(i);
                    if (lastArgStep.isTypeTest() && lastArgStep.getKindTest() == NodeKind.TEXT)
                        boolResult = true;
                }
            } else if (i == sizeLoc - 1 && j == sizeArg) {
                Step lastLocStep = loc.getStepNum(i);
                if (lastLocStep.isTypeTest() && lastLocStep.getKindTest() == NodeKind.TEXT)
                    boolResult = true;
            }
        }

    }

    public void visit(Step arg) throws XQueryException {
        boolResult = false;
        if (expression instanceof Step) {
            Step step = (Step) expression;
            if (step.getAxis() == arg.getAxis()) {
                expression = step.getExpression();
                arg.getExpression().accept(this);
            }
        }
    }

    public void visit(Variable arg) throws XQueryException {
        boolResult = false;
        if (expression instanceof Variable) {
            expression = ((Variable)expression).getVarName();
            arg.getVarName().accept(this);
        }
    }


    public void visit(QName arg) throws XQueryException {
        boolResult = arg.equals(expression);
    }

    public void visit(FunctionDATA arg) throws XQueryException {
        boolResult = false;
        arg.getArgument(0).accept(this);
    }
   
    public void visit(FunctionCall arg) throws XQueryException {
        boolResult = false;
        if (arg.getClass().getName().startsWith("org.xquark.xquery.parser.primitivefunctions.xsfunctions"))
            arg.getArgument(0).accept(this);
    }

    public void visit(XQueryExpression arg) throws XQueryException {
//        throw new XQueryException("Unsupported class for ExpressionComparatorVisitor : " + arg.getClass().getName());
    }

}