/*
 * This file belongs to the XQuare distribution. Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307. You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.parser;

import org.xquark.xquery.parser.primitivefunctions.fnfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionBOOLEAN;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionQNAME;
import org.xquark.xquery.parser.primitivefunctions.xqfunctions.FunctionXML_FRAGMENT;
import org.xquark.xquery.parser.primitivefunctions.xqjutilfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.FunctionSTRING;

public class DefaultParserVisitor implements ParserVisitor {

    private static final String RCSRevision = "$Revision: 1.30 $";

    private static final String RCSName = "$Name:  $";

    public void visit(AggregateFunctionCall arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(AttributeValuePair arg) throws XQueryException {
        visit((XQueryBinaryOperatorExpression) arg);
    }

    public void visit(BinOpANDExpression arg) throws XQueryException {
        visit((XQueryBooleanOperatorExpression) arg);
    }

    public void visit(BinOpORExpression arg) throws XQueryException {
        visit((XQueryBooleanOperatorExpression) arg);
    }

    public void visit(CaseClauseExpression arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(CastTreatExpression arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(CData arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(Document arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(Element arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    // public void visit(ExternalVariable arg) throws XQueryException {
    // visit((Variable) arg);
    // }

    public void visit(FLWRExpression arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(FunctionCall arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(FunctionDeclaration arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(InstanceOfExpression arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(InternalFunctionCall arg) throws XQueryException {
        visit((FunctionCall) arg);
    }

    public void visit(ITEExpression arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(ItemType arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(LibraryFunctionCall arg) throws XQueryException {
        visit((FunctionCall) arg);
    }

    public void visit(ListOpAFTERExpression arg) throws XQueryException {
        visit((XQueryListBinaryOperatorExpression) arg);
    }

    public void visit(ListOpArithExpression arg) throws XQueryException {
        visit((XQueryListBinaryOperatorExpression) arg);
    }

    public void visit(ListOpBEFOREExpression arg) throws XQueryException {
        visit((XQueryListBinaryOperatorExpression) arg);
    }

    public void visit(ListOpCompExpression arg) throws XQueryException {
        visit((XQueryListBinaryOperatorExpression) arg);
    }

    public void visit(ListOpEXCEPTExpression arg) throws XQueryException {
        visit((XQueryListBinaryOperatorExpression) arg);
    }

    public void visit(ListOpINTERSECTExpression arg) throws XQueryException {
        visit((XQueryListBinaryOperatorExpression) arg);
    }

    public void visit(ListOpUNIONExpression arg) throws XQueryException {
        visit((XQueryListBinaryOperatorExpression) arg);
    }

    public void visit(LocatedExpression arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(ElementTest arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(AttributeTest arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(SchemaElementTest arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(SchemaAttributeTest arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(NodeTest arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(PrimitiveFunctionCall arg) throws XQueryException {
        visit((FunctionCall) arg);
    }

    public void visit(QName arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(QuantifiedExpression arg) throws XQueryException {
        visit((XQueryListUnaryOperatorExpression) arg);
    }

    public void visit(RangeExpression arg) throws XQueryException {
        visit((XQueryListBinaryOperatorExpression) arg);
    }

    public void visit(InputFunctionCall arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(SequenceType arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(SortedExpression arg) throws XQueryException {
        visit((XQueryListUnaryOperatorExpression) arg);
    }

    public void visit(Step arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(TypeSwitchExpression arg) throws XQueryException {
        visit((XQueryBinaryOperatorExpression) arg);
    }

    public void visit(UnOpMinusExpression arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(ValidateExpression arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(Value arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(ValueBoolean arg) throws XQueryException {
        visit((Value) arg);
    }

    public void visit(ValueDecimal arg) throws XQueryException {
        visit((Value) arg);
    }

    public void visit(ValueDouble arg) throws XQueryException {
        visit((Value) arg);
    }

    public void visit(ValueFloat arg) throws XQueryException {
        visit((Value) arg);
    }

    public void visit(ValueInteger arg) throws XQueryException {
        visit((Value) arg);
    }

    public void visit(ValueString arg) throws XQueryException {
        visit((Value) arg);
    }

    public void visit(ValueText arg) throws XQueryException {
        visit((Value) arg);
    }

    public void visit(Variable arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(XQueryBinaryOperatorExpression arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(XMLComment arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(ComputedText arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(ComputedNamespace arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(XMLProcessingInstruction arg) throws XQueryException {
        visit((XQueryBinaryOperatorExpression) arg);
    }

    public void visit(XQueryBooleanOperatorExpression arg) throws XQueryException {
        visit((XQueryBinaryOperatorExpression) arg);
    }

    public void visit(ParenthesizedExpression arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(EnclosedExpression arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(XQueryExpression arg) throws XQueryException {
    }

    public void visit(XQueryExpressionSequence arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(XQueryListBinaryOperatorExpression arg) throws XQueryException {
        visit((XQueryBinaryOperatorExpression) arg);
    }

    public void visit(XQueryListUnaryOperatorExpression arg) throws XQueryException {
        visit((XQueryUnaryOperatorExpression) arg);
    }

    public void visit(XQueryUnaryOperatorExpression arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(XQueryModule arg) throws XQueryException {
        /* visit((XQueryUnit)arg); */}

    // primitive functions

    public void visit(FunctionABS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVG arg) throws XQueryException {
        visit((AggregateFunctionCall) arg);
    }

    public void visit(FunctionCEILING arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionCOLLECTION arg) throws XQueryException {
        visit((InputFunctionCall) arg);
    }

    public void visit(FunctionCONCAT arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionCONTAINS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionCOUNT arg) throws XQueryException {
        visit((AggregateFunctionCall) arg);
    }

    public void visit(FunctionCURRENT_DATE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionCURRENT_DATETIME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionCURRENT_TIME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionDATA arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionDATE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionDATETIME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionDECIMAL arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionDISTINCT_VALUES arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionDOC arg) throws XQueryException {
        visit((InputFunctionCall) arg);
    }

    public void visit(FunctionDOUBLE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionEMPTY arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionENDS_WITH arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionQNAME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionEXISTS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionFALSE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionFLOOR arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionMONTH_FROM_DATE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionYEAR_FROM_DATE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionINTEGER arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionLOCAL_NAME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionLOCAL_NAME_FROM_QNAME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionLOWER_CASE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionMATCHES arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionMAX arg) throws XQueryException {
        visit((AggregateFunctionCall) arg);
    }

    public void visit(FunctionMIN arg) throws XQueryException {
        visit((AggregateFunctionCall) arg);
    }

    public void visit(FunctionNAMESPACE_URI arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionNAMESPACE_URI_FROM_QNAME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionNOT arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionNUMBER arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(org.xquark.xquery.parser.primitivefunctions.xsfunctions.FunctionQNAME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionSTARTS_WITH arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionSTRING arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionSTRING_LENGTH arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionROUND arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionSUBSTRING arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionSUM arg) throws XQueryException {
        visit((AggregateFunctionCall) arg);
    }

    public void visit(FunctionTIME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionTRUE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionUPPER_CASE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionDEEP_EQUAL arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionALL_AVAILABLE_COLLECTION_URIS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionALL_AVAILABLE_FUNCTIONS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_COLLATION_URIS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_COLLECTION_URIS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_COLLECTIONS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_DOCUMENT_URIS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_DOCUMENTS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_FUNCTIONS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_MODULES arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_SCHEMAS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionAVAILABLE_STANDARD_FUNCTIONS arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionMETADATA arg) throws XQueryException {
        visit((InputFunctionCall) arg);
    }

    public void visit(FunctionDAYTIMEDURATION arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionIMPLICIT_TIMEZONE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionTIMEZONE_FROM_DATETIME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionTIMEZONE_FROM_DATE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionTIMEZONE_FROM_TIME arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionADJUST_DATETIME_TO_TIMEZONE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionADJUST_DATE_TO_TIMEZONE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionADJUST_TIME_TO_TIMEZONE arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(FunctionBASE_URI arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    // public void visit(FunctionBOOLEAN arg)throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionCODEPOINTS_TO_STRING arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionCODEPOINT_COMPARE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionCOMPARE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionCONCATENATE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionDAYS_FROM_DURATION arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionDAY_FROM_DATE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionDAY_FROM_DATETIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionDEFAULT_COLLATION arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionDOCUMENT_URI arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionDOC_AVAILABLE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionENCODE_FOR_URI arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionERROR arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionESCAPE_HTML_URI arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionEXACTLY_ONE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionHOURS_FROM_DATETIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionHOURS_FROM_DURATION arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionHOURS_FROM_TIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionID arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionIDREF arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionINDEX_OF arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionINSERT_BEFORE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionIN_SCOPE_PREFIXES arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionIRI_TO_URI arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionLANG arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionLAST arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // public void visit(FunctionLOCAL_NAME_FROM_QNAME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionLOWER_CASE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMATCHES arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMAX arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMIN arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMINUTES_FROM_DATETIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMINUTES_FROM_DURATION arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMINUTES_FROM_TIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMONTHS_FROM_DURATION arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMONTH_FROM_DATE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionMONTH_FROM_DATETIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNAME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNAMESPACE_URI arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNAMESPACE_URI_FOR_PREFIX arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNAMESPACE_URI_FROM_QNAME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNILLED arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNODE_NAME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNORMALISE_SPACE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNORMALIZE_UNICODE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNOT arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionNUMBER arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionONE_OR_MORE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionPOSITION arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionPREFIX_FROM_QNAME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionQNAME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionREMOVE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionREPLACE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionRESOLVE_QNAME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionRESOLVE_URI arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionREVERSE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionROOT arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionROUND arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionROUND_HALF_TO_EVEN arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSECONDS_FROM_DATETIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSECONDS_FROM_DURATION arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSECONDS_FROM_TIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSTARTS_WITH arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSTATIC_BASE_URI arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSTRING arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSTRING_JOIN arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSTRING_LENGTH arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSTRING_TO_CODEPOINTS arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSUBSEQUENCE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSUBSTRING arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSUBSTRING_AFTER arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSUBSTRING_BEFORE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionSUM arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionTOKENIZE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionTRACE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionTRANSLATE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionTRUE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionUNORDERED arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionUPPER_CASE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionYEARS_FROM_DURATION arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionYEAR_FROM_DATE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionYEAR_FROM_DATETIME arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }
    // FunctionZERO_OR_MORE arg) throws XQueryException {visit((PrimitiveFunctionCall) arg); }

    public void visit(FunctionXML_FRAGMENT arg) throws XQueryException {
        visit((PrimitiveFunctionCall) arg);
    }

    public void visit(ImportModule arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

    public void visit(ImportSchema arg) throws XQueryException {
        visit((XQueryExpression) arg);
    }

}