/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.reconstruction;

import java.io.IOException;
import java.io.StringReader;
import java.util.*;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.XMLFilterImpl;
import org.xquark.datastructure.EvaluationVisitor;
import org.xquark.schema.ElementDeclaration;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.PSVInfoSetProvider;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.util.*;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.metadata.DynamicContext;
import org.xquark.xquery.normalize.GetBoundVariablesVisitor;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionDATA;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionQNAME;
import org.xquark.xquery.parser.primitivefunctions.xqfunctions.FunctionXML_FRAGMENT;
import org.xquark.xquery.typing.QType;
import org.xquark.xquery.typing.QTypeElement;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;
import org.xquark.xquery.xdbc.XResultSetImpl;

public class ReconstructionVisitor extends DefaultParserVisitor {

    private static final String RCSRevision = "$Revision: 1.41 $";

    private static final String RCSName = "$Name:  $";

    private final String EMPTY_STRING = "";

    // private final String WHITE_SPACE = " ";

    static final private String PREFIXROOT = "xqns";

    // private ContentHandler contenthandler = null;
    // private LexicalHandler lexicalhandler = null;
    // private ErrorHandler errorhandler = null;

    private ReconstructionHandler reconstructionHandler = null;

    private XResultSetImpl resultSet = null;

    private XDBCResultSetInterface xdbcResultSet = null;

    private EvaluationVisitor evalVisitor = null;

    private AttributesImpl listAttributes = null;

    private Object[] lastIdentifiers = null;

    private Object[] currentIdentifiers = null;

    private Object[] nextIdentifiers = null;

    private boolean noNext = false;

    private boolean noReconstructor = false;

    private boolean hasidentifiers = false;

    // private boolean hasPrefixMapping = false;

    // private boolean outsidePutPrefixMapping = false;

    private NamespaceContextStack declaredDeclarations = null;

    private NamespaceContextStack reconstructionDeclarations = new NamespaceContextStack();

    // private XMLReader reader = null;
    private FragmentFilter filter = null;

    private ValidationContextProvider vcp = new ValidationContextProvider() {
        public String getNamespaceURI(String prefix) {
            return reconstructionDeclarations.getNamespaceURI(prefix);
        }

        public String getPrefix(String uri) {
            return reconstructionDeclarations.getPrefix(uri);
        }

        public String getDocumentBase() {
            return null;
        }

        public Map getNotationDeclarations() {
            return null;
        }
    };

    private int size = -1;

    private ArrayList varList = null;

    private int lastSkolemIDsSize = -1;

    private ArrayList lastDependIDs = null;

    private boolean evalNeeded = false;

    private String evalResultNamespace = null;

    private String evalResultValue = null;

    private boolean root = true;

    private boolean noResultSet = false;

    private boolean noReInit = false;

    private ArrayList skolemIDs = null;

    // private DynamicContext dc = null;

    private PSVInfoSetProvider psvisp = null;

    private class ReconstructionHandler extends NamespaceContextHandler {
        private final String WHITE_SPACE = " ";

        private boolean prevIsOnlyChars = false;

        private boolean isOnlyChars = false;

        public boolean IsOnlyChars() {
            return isOnlyChars;
        }

        public void setIsOnlyChar(boolean isOnlyChars) {
            this.isOnlyChars = isOnlyChars;
        }

        public void setPrevIsOnlyChar(boolean prevIsOnlyChars) {
            this.prevIsOnlyChars = prevIsOnlyChars;
        }

        public void characters(char[] values, int start, int length) throws SAXException {
            if (prevIsOnlyChars) {
                super.characters(WHITE_SPACE.toCharArray(), 0, 1);
            }
            if (!isOnlyChars)
                isOnlyChars = true;
            super.characters(values, start, length);
        }

        public void startPrefixMapping(String prefix, String uri) throws SAXException {
            String ns = getNamespaceURI(prefix);
            if (ns == null || !ns.equals(uri))
                super.startPrefixMapping(prefix, uri);
        }

        public void endPrefixMapping(String prefix) throws SAXException {
            if (getDeclaredPrefixes().contains(prefix))
                super.endPrefixMapping(prefix);
        }

        public void startElement(String uri, String localname, String qname, Attributes attributes) throws SAXException {
            // contextStack.pushContext();
            checkPrefix(uri, declaredDeclarations);
            for (int i = 0; i < attributes.getLength(); i++) {
                checkPrefix(attributes.getURI(i), declaredDeclarations);
            }
            super.startElement(uri, localname, qname, attributes);
            if (isOnlyChars)
                isOnlyChars = false;
            if (prevIsOnlyChars)
                prevIsOnlyChars = false;
        }

        public String checkPrefix(String uri, NamespaceContextStack contextStack) throws SAXException {
            if (uri == null || uri.length() == 0)
                return null;
            String prefix = getPrefix(uri);
            if (prefix == null) {
                prefix = (contextStack == null) ? null : contextStack.getPrefix(uri);
                if (prefix == null) {
                    for (int i = 1;; i++) {
                        String genPrefix = PREFIXROOT + i;
                        if (getNamespaceURI(genPrefix) == null) {
                            prefix = genPrefix;
                            break;
                        }
                    }
                }
                this.startPrefixMapping(prefix, uri);
            }
            return prefix;
        }

        public void processingInstruction(String str, String str1) throws SAXException {
            super.processingInstruction(str, str1);
            if (isOnlyChars)
                isOnlyChars = false;
            if (prevIsOnlyChars)
                prevIsOnlyChars = false;
        }

        public void endElement(String uri, String localname, String qname) throws SAXException {
            super.endElement(uri, localname, qname);
            List list = getDeclaredPrefixes();
            for (int i = 0; i < list.size(); i++)
                this.endPrefixMapping((String) list.get(i));
            // contextStack.popContext();
            if (isOnlyChars)
                isOnlyChars = false;
            if (prevIsOnlyChars)
                prevIsOnlyChars = false;
        }
    }

    private class FragmentFilter extends DefaultXMLFilter {
        ElementDeclaration eltDecl = null;

        private int count = 0;

        private boolean newPrefix = false;

        public void setQType(QType qtype) {
            if (qtype != null && qtype instanceof QTypeElement)
                eltDecl = ((QTypeElement) qtype).getElementDeclaration();
        }

        public void startDocument() throws SAXException {
            count = 0;
            newPrefix = false;
        }

        public void endDocument() throws SAXException {
        }

        public void startPrefixMapping(String prefix, String uri) throws SAXException {
            String ns = ((ReconstructionHandler) getContentHandler()).getNamespaceURI(prefix);
            if (ns == null || !ns.equals(uri))
                super.startPrefixMapping(prefix, uri);
        }

        public void endPrefixMapping(String prefix) throws SAXException {
            if (((ReconstructionHandler) getContentHandler()).getDeclaredPrefixes().contains(prefix))
                super.endPrefixMapping(prefix);
        }

        public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
            if (count == 0 && eltDecl != null) {
                ReconstructionHandler handler = (ReconstructionHandler) getContentHandler();
                String prefix = handler.getPrefix(eltDecl.getNamespace());
                if (prefix == null) {
                    for (int i = 1;; i++) {
                        String genPrefix = PREFIXROOT + i;
                        if (handler.getNamespaceURI(genPrefix) == null) {
                            prefix = genPrefix;
                            break;
                        }
                    }
                    newPrefix = true;
                    super.startPrefixMapping(prefix, eltDecl.getNamespace());
                }
                super.startElement(eltDecl.getNamespace(), eltDecl.getName(), null, atts);
            } else
                super.startElement(uri, localName, qName, atts);
            count++;
        }

        public void endElement(String uri, String localName, String qName) throws SAXException {
            count--;
            if (count == 0 && eltDecl != null) {
                super.endElement(eltDecl.getNamespace(), eltDecl.getName(), null);
                if (newPrefix) {
                    ReconstructionHandler handler = (ReconstructionHandler) getContentHandler();
                    String prefix = handler.getPrefix(uri);
                    if (prefix != null)
                        super.endPrefixMapping(prefix);
                }
            } else
                super.endElement(uri, localName, qName);
        }
    }

    public ReconstructionVisitor(SchemaManager sm, DynamicContext dc, XResultSetImpl resultSet, ArrayList varList) throws XQueryException {
        this.reconstructionHandler = new ReconstructionHandler();
        try {
            SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setNamespaceAware(true);
            XMLReader reader = factory.newSAXParser().getXMLReader();
            filter = new FragmentFilter();
            filter.setParent(reader);
            filter.setContentHandler(reconstructionHandler);
            filter.setErrorHandler(reconstructionHandler);
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        } catch (ParserConfigurationException e) {
            throw new XQueryException(e.getMessage());
        }
        this.resultSet = resultSet;
        this.xdbcResultSet = resultSet.getXDBCResultSet();
        if (xdbcResultSet != null) {
            xdbcResultSet.setContentHandler(reconstructionHandler);
            xdbcResultSet.setLexicalHandler(reconstructionHandler);
        }
        psvisp = resultSet.getPSVInfoSetProvider();
        this.varList = varList;
        size = varList.size();
        currentIdentifiers = new Object[size];
        nextIdentifiers = new Object[size];
        lastIdentifiers = new Object[size];
        listAttributes = new AttributesImpl();
        evalVisitor = new EvaluationVisitor(sm, dc, xdbcResultSet, vcp);
    }

    public void reInit() {
        if (!noReInit)
            noNext = false;
        else
            noReInit = false;
        root = true;
        // hasPrefixMapping = false;
        hasidentifiers = false;
    }

    private void setIdentifiers(boolean force) throws XQueryException {
        if (!force && hasidentifiers)
            return;
        try {
            if (xdbcResultSet.hasNext())
                xdbcResultSet.getNextIdentifiers(nextIdentifiers);
            if (!this.noResultSet)
                xdbcResultSet.getIdentifiers(currentIdentifiers);
            hasidentifiers = true;
        } catch (XMLDBCException xmldbce) {
            throw new XQueryException(xmldbce.getMessage());
        }
    }

    public void setContentHandler(ContentHandler handler) {
        reconstructionHandler.setContentHandler(handler);
        if (xdbcResultSet != null)
            xdbcResultSet.setContentHandler(reconstructionHandler);
    }

    /**
     * @see XDBCResultSetInterface#setLexicalHandler
     */
    public void setLexicalHandler(LexicalHandler handler) {
        reconstructionHandler.setLexicalHandler(handler);
        if (xdbcResultSet != null)
            xdbcResultSet.setLexicalHandler(reconstructionHandler);
    }

    /**
     * @see XDBCResultSetInterface#setErrorHandler
     */
    public void setErrorHandler(ErrorHandler handler) {
        // errorhandler = handler;
    }

    public XResultSetImpl getResultSet() {
        return resultSet;
    }

    /*
     * public ContentHandler getContentHandler() { return reconstructionHandler.getContentHandler(); } public LexicalHandler getLexicalHandler() { return reconstructionHandler.getLexicalHandler(); } public ErrorHandler getErrorHandler() { return errorhandler; }
     */
    public boolean getNoReconstructor() {
        return noReconstructor;
    }

    public boolean getNoNext() {
        return noNext;
    }

    public boolean getNoResultSet() {
        return noResultSet;
    }

    public void setNoResultSet(boolean noResultSet) {
        this.noResultSet = noResultSet;
    }

    public void setNamespaceDecl(NamespaceContextStack declarations) {
        // hasPrefixMapping = true;
        this.declaredDeclarations = declarations;
    }

    // visitor ...

    // public void visit(AggregateFunctionCall arg) throws XQueryException

    private void outputAttributeValuePair(AttributeValuePair arg) throws XQueryException {
        // if (arg.isXmlns())
        // return;
        if (arg.isConditionnal()) {
            evalVisitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
            arg.getAttributeValue().accept(evalVisitor);
            if (!evalVisitor.getVerdict()) {
                // reset key variables
                skolemIDs = arg.getSkolemIDs();
                if (skolemIDs != null)
                    lastSkolemIDsSize = skolemIDs.size();
                return;
            }
        }

        String nameSpaceAttName = EMPTY_STRING;
        String localNameAttName = EMPTY_STRING;
        String attValue = EMPTY_STRING;
        XQueryExpression attrName = arg.getAttributeName();
        evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
        attrName.accept(evalVisitor);
        Comparable comp = evalVisitor.getResValue();
        String namePrefix = null;
        if (comp != null) {
            // this is not nice but for now it works fine
            String compStr = comp.toString();
            if (compStr.startsWith("{")) {
                int ind = compStr.lastIndexOf('}');
                if (ind == -1)
                    localNameAttName = compStr;
                else {
                    nameSpaceAttName = compStr.substring(1, ind);
                    localNameAttName = compStr.substring(ind + 1);
                }
            } else {
                localNameAttName = compStr;
                if (!arg.isXmlns()) {
                    int ind = compStr.lastIndexOf(':');
                    if (ind == -1)
                        localNameAttName = compStr;
                    else {
                        namePrefix = compStr.substring(0, ind);
                        nameSpaceAttName = reconstructionDeclarations.getNamespaceURI(namePrefix);
                        if (nameSpaceAttName == null) {
                            if (declaredDeclarations != null)
                                nameSpaceAttName = declaredDeclarations.getNamespaceURI(namePrefix);
                            if (nameSpaceAttName != null)
                                declarePrefix(namePrefix, nameSpaceAttName);
                        }
                        localNameAttName = compStr.substring(ind + 1);
                    }
                }
            }
        }

        XQueryExpression attrValue = arg.getAttributeValue();
        evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
        attrValue.accept(evalVisitor);
        comp = evalVisitor.getResValue();
        if (comp != null) {
            // this is not nice but for now it works fine
            attValue = comp.toString();
            // TODO ANNUAIRE verify that value is QName before testing first character
            // the column metadata should be set to QName in the configuration file?
            // if (attrValue.getQType().isQName() && attValue.startsWith("{")) {
            if (attValue.startsWith("{")) {
                int ind = attValue.lastIndexOf('}');
                if (ind != -1) {
                    String namespace = attValue.substring(1, ind);
                    String value = attValue.substring(ind + 1);
                    String valPrefix = reconstructionDeclarations.getPrefix(namespace);
                    if (valPrefix == null) {
                        try {
                            valPrefix = reconstructionHandler.checkPrefix(namespace, reconstructionDeclarations);
                        } catch (SAXException e) {
                            throw new XQueryException("Could not find prefix.", e);
                        }
                    }
                    // outsidePutPrefixMapping = true;
                    attValue = valPrefix + ":" + value;
                }
            }
        }
        if (arg.isXmlns()) {
            declarePrefix(localNameAttName, attValue);
            localNameAttName = "xmlns:" + localNameAttName;
        }

        // test if attribute already exists
        if (listAttributes.getValue(nameSpaceAttName, localNameAttName) != null)
            throw new ReconstructionException("An element cannot have several attributes with the same name.");
        // add attribute
        if (!arg.isXmlns())
            listAttributes.addAttribute((nameSpaceAttName == null ? "" : nameSpaceAttName), localNameAttName, (namePrefix == null ? "" : namePrefix + ":") + localNameAttName, "STRING", attValue);

        // reset key variables
        skolemIDs = arg.getSkolemIDs();
        if (skolemIDs != null)
            lastSkolemIDsSize = skolemIDs.size();
    }

    public void visit(AttributeValuePair arg) throws XQueryException {
        // if (arg.isXmlns())
        // return;
        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        skolemIDs = arg.getSkolemIDs();
        if (skolemIDs != null)
            lastSkolemIDsSize = skolemIDs.size();
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputAttributeValuePair(arg);
                    root = previousRoot;
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputAttributeValuePair(arg);
                    root = previousRoot;
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                // WRITE SONS
                boolean previousRoot = root;
                root = false;
                outputAttributeValuePair(arg);
                root = previousRoot;
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputAttributeValuePair(arg);
                        root = previousRoot;
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputAttributeValuePair(arg);
                    root = previousRoot;
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;
                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputAttributeValuePair(arg);
                        root = previousRoot;
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputAttributeValuePair(arg);
                        root = previousRoot;
                        if (!hasNext())
                            break;
                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        noReconstructor = false;
    }

    // public void visit(BinOpANDExpression arg) throws XQueryException;
    // public void visit(BinOpORExpression arg) throws XQueryException;
    // public void visit(CastTreatExpression arg) throws XQueryException;

    public void visit(CData arg) throws XQueryException {
        try {
            reconstructionHandler.startCDATA();
            characters(arg.getCData().toCharArray(), 0, arg.getCData().length());
            reconstructionHandler.endCDATA();
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    // public void visit(ContextDeclaration arg) throws XQueryException;

    private void outputComputedText(ComputedText arg) throws XQueryException {
        evalNeeded = true;
        arg.getExpression().accept(this);
        evalNeeded = false;
        if (evalResultValue != null) {
            characters(evalResultValue.toCharArray(), 0, evalResultValue.length());
            reconstructionHandler.setPrevIsOnlyChar(true);
            evalResultValue = null;
        }
        // reset key variables
        skolemIDs = arg.getSkolemIDs();
        if (skolemIDs != null)
            lastSkolemIDsSize = skolemIDs.size();
    }

    public void visit(ComputedText arg) throws XQueryException {

        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        skolemIDs = arg.getSkolemIDs();
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputComputedText(arg);
                    root = previousRoot;
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputComputedText(arg);
                    root = previousRoot;
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                // WRITE SONS
                boolean previousRoot = root;
                root = false;
                outputComputedText(arg);
                root = previousRoot;
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputComputedText(arg);
                        root = previousRoot;
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputComputedText(arg);
                    root = previousRoot;
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;
                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputComputedText(arg);
                        root = previousRoot;
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        arg.getExpression().setRoot(true);
                        outputComputedText(arg);
                        root = previousRoot;
                        if (!hasNext())
                            break;
                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        noReconstructor = false;
    }

    // public void visit(Dereference arg) throws XQueryException;

    private void outputDocument(Document arg) throws XQueryException {
        arg.getExpression().accept(this);
        // reset key variables
        skolemIDs = arg.getSkolemIDs();
        if (skolemIDs != null)
            lastSkolemIDsSize = skolemIDs.size();
    }

    public void visit(Document arg) throws XQueryException {

        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        skolemIDs = arg.getSkolemIDs();
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    try {
                        reconstructionHandler.startDocument();
                    } catch (SAXException e) {
                        throw new XQueryException(e.getMessage());
                    }
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputDocument(arg);
                    root = previousRoot;
                    // WRITE STOP
                    try {
                        reconstructionHandler.endDocument();
                    } catch (SAXException e) {
                        throw new XQueryException(e.getMessage());
                    }
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // case of namespace in top element
                    try {
                        reconstructionHandler.startDocument();
                    } catch (SAXException e) {
                        throw new XQueryException(e.getMessage());
                    }
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputDocument(arg);
                    root = previousRoot;
                    // WRITE STOP
                    try {
                        reconstructionHandler.endDocument();
                    } catch (SAXException e) {
                        throw new XQueryException(e.getMessage());
                    }
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                try {
                    reconstructionHandler.startDocument();
                } catch (SAXException e) {
                    throw new XQueryException(e.getMessage());
                }
                // WRITE SONS
                boolean previousRoot = root;
                root = false;
                outputDocument(arg);
                root = previousRoot;
                // WRITE STOP
                try {
                    reconstructionHandler.endDocument();
                } catch (SAXException e) {
                    throw new XQueryException(e.getMessage());
                }
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        try {
                            reconstructionHandler.startDocument();
                        } catch (SAXException e) {
                            throw new XQueryException(e.getMessage());
                        }
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputDocument(arg);
                        root = previousRoot;
                        // WRITE STOP
                        try {
                            reconstructionHandler.endDocument();
                        } catch (SAXException e) {
                            throw new XQueryException(e.getMessage());
                        }
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    try {
                        reconstructionHandler.startDocument();
                    } catch (SAXException e) {
                        throw new XQueryException(e.getMessage());
                    }
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputDocument(arg);
                    root = previousRoot;
                    // WRITE STOP
                    try {
                        reconstructionHandler.endDocument();
                    } catch (SAXException e) {
                        throw new XQueryException(e.getMessage());
                    }
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;

                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        try {
                            reconstructionHandler.startDocument();
                        } catch (SAXException e) {
                            throw new XQueryException(e.getMessage());
                        }
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputDocument(arg);
                        root = previousRoot;
                        // WRITE STOP
                        try {
                            reconstructionHandler.endDocument();
                        } catch (SAXException e) {
                            throw new XQueryException(e.getMessage());
                        }
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        try {
                            reconstructionHandler.startDocument();
                        } catch (SAXException e) {
                            throw new XQueryException(e.getMessage());
                        }
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        arg.getExpression().setRoot(true);
                        outputDocument(arg);
                        root = previousRoot;
                        // WRITE STOP
                        try {
                            reconstructionHandler.endDocument();
                        } catch (SAXException e) {
                            throw new XQueryException(e.getMessage());
                        }
                        if (!hasNext())
                            break;
                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        noReconstructor = false;
    }

    private GetBoundVariablesVisitor gbvv = new GetBoundVariablesVisitor(true, false, false);

    public void visit(Element arg) throws XQueryException {

        if (evalNeeded) {
            evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
            arg.accept(evalVisitor);
            evalResultNamespace = EMPTY_STRING;
            Comparable comp = evalVisitor.getResValue();
            if (comp == null)
                evalResultValue = EMPTY_STRING;
            else {
                if (evalResultValue != null && evalResultValue.length() > 0)
                    evalResultValue += " " + comp.toString();
                else
                    evalResultValue = comp.toString();
            }
            return;
        }

        // pushContext();

        // HashMap localPrefixes = arg.getLocalPrefixes();
        // if (localPrefixes != null) {
        // Iterator it = localPrefixes.keySet().iterator();
        // String prefix = null;
        // while (it.hasNext()) {
        // prefix = (String) it.next();
        // declarePrefix(prefix, (String) localPrefixes.get(prefix));
        // }
        // }
        // if (!hasPrefixMapping && localPrefixes != null)
        // hasPrefixMapping = true;
        boolean isSkolemID = false;
        boolean isDependID = false;
        // boolean putPrefixMapping = false;
        noReconstructor = true;
        skolemIDs = arg.getSkolemIDs();
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        XQueryExpression startTag = arg.getStartTag();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    pushContext(arg);
                    evalNeeded = true;
                    startTag.accept(this);
                    evalNeeded = false;
                    String nameSpace = evalResultNamespace;
                    String localName = evalResultValue;
                    evalResultNamespace = null;
                    evalResultValue = null;
                    // special for computed namespaces
                    if (arg.getComputedNSNum() > 0) {
                        for (int i = 0; i < arg.getComputedNSNum(); i++) {
                            ComputedNamespace compNS = (ComputedNamespace) ((EnclosedExpression) arg.getSubExpressions().get(i)).getExpression();
                            evalNeeded = true;
                            compNS.getExpression().accept(this);
                            evalNeeded = false;
                            declarePrefix(compNS.getNcname(), evalResultValue == null ? "" : evalResultValue);
                        }
                        // hasPrefixMapping = true;
                    }
                    // prepare attributes
                    prepareAttributes(arg);
                    // case of attributes (to be evaluated in sons) to add at <start_element>
                    buildAttributes(arg);
                    if (nameSpace != EMPTY_STRING) {
                        String prefix = reconstructionDeclarations.getPrefix(nameSpace);
                        if (prefix == null) {
                            if (declaredDeclarations != null)
                                prefix = declaredDeclarations.getPrefix(nameSpace);
                            if (prefix != null)
                                declarePrefix(prefix, nameSpace == null ? "" : nameSpace);
                        }
                    }
                    startPrefixMapping(root);
                    // WRITE START
                    startElement(nameSpace, localName, listAttributes);
                    // WRITE SONS
                    ArrayList subExpressions = arg.getSubExpressions();
                    if (subExpressions != null) {
                        boolean previousRoot = root;
                        root = false;
                        reconstructionHandler.setPrevIsOnlyChar(false);
                        for (int i = arg.getComputedNSNum() + arg.getComputedAttNum(); i < subExpressions.size(); i++) {
                            XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
                            expr.accept(this);
                            reconstructionHandler.setPrevIsOnlyChar(reconstructionHandler.IsOnlyChars());
                            this.skolemIDs = arg.getSkolemIDs();
                            if (skolemIDs != null)
                                lastSkolemIDsSize = skolemIDs.size();
                        }
                        root = previousRoot;
                    }
                    // WRITE STOP
                    endElement(nameSpace, localName);
                    endPrefixMapping(root);
                    popContext(arg);
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                noNext = false;
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    pushContext(arg);
                    // case of namespace in top element
                    evalNeeded = true;
                    startTag.accept(this);
                    evalNeeded = false;
                    String nameSpace = evalResultNamespace;
                    String localName = evalResultValue;
                    evalResultNamespace = null;
                    evalResultValue = null;
                    // special for computed namespaces
                    if (arg.getComputedNSNum() > 0) {
                        for (int i = 0; i < arg.getComputedNSNum(); i++) {
                            ComputedNamespace compNS = (ComputedNamespace) ((EnclosedExpression) arg.getSubExpressions().get(i)).getExpression();
                            evalNeeded = true;
                            compNS.getExpression().accept(this);
                            evalNeeded = false;
                            declarePrefix(compNS.getNcname(), evalResultValue == null ? "" : evalResultValue);
                        }
                        // hasPrefixMapping = true;
                    }
                    // prepare attributes
                    prepareAttributes(arg);
                    // case of attributes (to be evaluated in sons) to add at <start_element>
                    buildAttributes(arg);
                    // start prefix mapping
                    if (nameSpace != EMPTY_STRING) {
                        String prefix = reconstructionDeclarations.getPrefix(nameSpace);
                        if (prefix == null) {
                            if (declaredDeclarations != null)
                                prefix = declaredDeclarations.getPrefix(nameSpace);
                            if (prefix != null)
                                declarePrefix(prefix, nameSpace == null ? "" : nameSpace);
                        }
                    }
                    startPrefixMapping(root);
                    // WRITE START
                    startElement(nameSpace, localName, listAttributes);
                    // set lastSkolemIDsSize
                    this.skolemIDs = arg.getSkolemIDs();
                    if (skolemIDs != null)
                        lastSkolemIDsSize = skolemIDs.size();
                    // WRITE SONS
                    ArrayList subExpressions = arg.getSubExpressions();
                    if (subExpressions != null) {
                        boolean previousRoot = root;
                        root = false;
                        reconstructionHandler.setPrevIsOnlyChar(false);
                        for (int i = arg.getComputedAttNum() + arg.getComputedNSNum(); i < subExpressions.size(); i++) {
                            XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
                            expr.accept(this);
                            reconstructionHandler.setPrevIsOnlyChar(reconstructionHandler.IsOnlyChars());
                            this.skolemIDs = arg.getSkolemIDs();
                            if (skolemIDs != null)
                                lastSkolemIDsSize = skolemIDs.size();
                        }
                        root = previousRoot;
                    }
                    // WRITE STOP
                    endElement(nameSpace, localName);
                    endPrefixMapping(root);
                    popContext(arg);
                    if (!hasNext()) {
                        break;
                    }
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext()) {
                            break;
                        }
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                pushContext(arg);
                evalNeeded = true;
                startTag.accept(this);
                evalNeeded = false;
                String nameSpace = evalResultNamespace;
                String localName = evalResultValue;
                evalResultNamespace = null;
                evalResultValue = null;
                // special for computed namespaces
                if (arg.getComputedNSNum() > 0) {
                    for (int i = 0; i < arg.getComputedNSNum(); i++) {
                        ComputedNamespace compNS = (ComputedNamespace) ((EnclosedExpression) arg.getSubExpressions().get(i)).getExpression();
                        evalNeeded = true;
                        compNS.getExpression().accept(this);
                        evalNeeded = false;
                        declarePrefix(compNS.getNcname(), evalResultValue == null ? "" : evalResultValue);
                    }
                    // hasPrefixMapping = true;
                }
                // prepare attributes
                prepareAttributes(arg);
                // case of attributes (to be evaluated in sons) to add at <start_element>
                buildAttributes(arg);
                // case of namespace in top element
                if (nameSpace != EMPTY_STRING) {
                    String prefix = reconstructionDeclarations.getPrefix(nameSpace);
                    if (prefix == null) {
                        if (declaredDeclarations != null)
                            prefix = declaredDeclarations.getPrefix(nameSpace);
                        if (prefix != null)
                            declarePrefix(prefix, nameSpace == null ? "" : nameSpace);
                    }
                }
                startPrefixMapping(root);
                // WRITE START
                startElement(nameSpace, localName, listAttributes);
                // WRITE SONS
                ArrayList subExpressions = arg.getSubExpressions();
                if (subExpressions != null) {
                    boolean previousRoot = root;
                    root = false;
                    reconstructionHandler.setPrevIsOnlyChar(false);
                    for (int i = arg.getComputedAttNum() + arg.getComputedNSNum(); i < subExpressions.size(); i++) {
                        XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
                        expr.accept(this);
                        reconstructionHandler.setPrevIsOnlyChar(reconstructionHandler.IsOnlyChars());
                        this.skolemIDs = arg.getSkolemIDs();
                    }
                    root = previousRoot;
                }
                // WRITE STOP
                endElement(nameSpace, localName);
                endPrefixMapping(root);
                popContext(arg);
            }
        }
        // there is at least a skolemID
        else {
            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        pushContext(arg);
                        evalNeeded = true;
                        startTag.accept(this);
                        evalNeeded = false;
                        String nameSpace = evalResultNamespace;
                        String localName = evalResultValue;
                        evalResultNamespace = null;
                        evalResultValue = null;
                        // special for computed namespaces
                        if (arg.getComputedNSNum() > 0) {
                            for (int i = 0; i < arg.getComputedNSNum(); i++) {
                                ComputedNamespace compNS = (ComputedNamespace) ((EnclosedExpression) arg.getSubExpressions().get(i)).getExpression();
                                evalNeeded = true;
                                compNS.getExpression().accept(this);
                                evalNeeded = false;
                                declarePrefix(compNS.getNcname(), evalResultValue == null ? "" : evalResultValue);
                            }
                            // hasPrefixMapping = true;
                        }
                        // prepare attributes
                        prepareAttributes(arg);
                        // case of attributes (to be evaluated in sons) to add at <start_element>
                        buildAttributes(arg);
                        // case of namespace in top element
                        if (nameSpace != EMPTY_STRING) {
                            String prefix = reconstructionDeclarations.getPrefix(nameSpace);
                            if (prefix == null) {
                                if (declaredDeclarations != null)
                                    prefix = declaredDeclarations.getPrefix(nameSpace);
                                if (prefix != null)
                                    declarePrefix(prefix, nameSpace == null ? "" : nameSpace);
                            }
                        }
                        startPrefixMapping(root);
                        // WRITE START
                        startElement(nameSpace, localName, listAttributes);
                        // set lastSkolemIDsSize
                        this.skolemIDs = arg.getSkolemIDs();
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        // WRITE SONS
                        ArrayList subExpressions = arg.getSubExpressions();
                        if (subExpressions != null) {
                            boolean previousRoot = root;
                            root = false;
                            reconstructionHandler.setPrevIsOnlyChar(false);
                            for (int i = arg.getComputedAttNum() + arg.getComputedNSNum(); i < subExpressions.size(); i++) {
                                XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
                                expr.accept(this);
                                reconstructionHandler.setPrevIsOnlyChar(reconstructionHandler.IsOnlyChars());
                                this.skolemIDs = arg.getSkolemIDs();
                                if (skolemIDs != null)
                                    lastSkolemIDsSize = skolemIDs.size();
                            }
                            root = previousRoot;
                        }
                        // WRITE STOP
                        endElement(nameSpace, localName);
                        endPrefixMapping(root);
                        popContext(arg);
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else {
                            break;
                        }
                    }
                } else {
                    pushContext(arg);
                    evalNeeded = true;
                    startTag.accept(this);
                    evalNeeded = false;
                    String nameSpace = evalResultNamespace;
                    String localName = evalResultValue;
                    evalResultNamespace = null;
                    evalResultValue = null;
                    // special for computed namespaces
                    if (arg.getComputedNSNum() > 0) {
                        for (int i = 0; i < arg.getComputedNSNum(); i++) {
                            ComputedNamespace compNS = (ComputedNamespace) ((EnclosedExpression) arg.getSubExpressions().get(i)).getExpression();
                            evalNeeded = true;
                            compNS.getExpression().accept(this);
                            evalNeeded = false;
                            declarePrefix(compNS.getNcname(), evalResultValue == null ? "" : evalResultValue);
                        }
                        // hasPrefixMapping = true;
                    }
                    // prepare attributes
                    prepareAttributes(arg);
                    // case of attributes (to be evaluated in sons) to add at <start_element>
                    buildAttributes(arg);
                    // case of namespace in top element
                    if (nameSpace != EMPTY_STRING) {
                        String prefix = reconstructionDeclarations.getPrefix(nameSpace);
                        if (prefix == null) {
                            if (declaredDeclarations != null)
                                prefix = declaredDeclarations.getPrefix(nameSpace);
                            if (prefix != null)
                                declarePrefix(prefix, nameSpace == null ? "" : nameSpace);
                        }
                    }
                    startPrefixMapping(root);
                    // WRITE START
                    startElement(nameSpace, localName, listAttributes);
                    // set lastSkolemIDsSize
                    this.skolemIDs = arg.getSkolemIDs();
                    if (skolemIDs != null)
                        lastSkolemIDsSize = skolemIDs.size();
                    // WRITE SONS
                    ArrayList subExpressions = arg.getSubExpressions();
                    if (subExpressions != null) {
                        boolean previousRoot = root;
                        root = false;
                        reconstructionHandler.setPrevIsOnlyChar(false);
                        for (int i = arg.getComputedAttNum() + arg.getComputedNSNum(); i < subExpressions.size(); i++) {
                            XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
                            expr.accept(this);
                            reconstructionHandler.setPrevIsOnlyChar(reconstructionHandler.IsOnlyChars());
                            this.skolemIDs = arg.getSkolemIDs();
                            if (skolemIDs != null)
                                lastSkolemIDsSize = skolemIDs.size();
                        }
                        root = previousRoot;
                    }
                    // WRITE STOP
                    endElement(nameSpace, localName);
                    endPrefixMapping(root);
                    popContext(arg);
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;

                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        // popContext(arg);
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        pushContext(arg);
                        evalNeeded = true;
                        startTag.accept(this);
                        evalNeeded = false;
                        String nameSpace = evalResultNamespace;
                        String localName = evalResultValue;
                        evalResultNamespace = null;
                        evalResultValue = null;
                        // special for computed namespaces
                        if (arg.getComputedNSNum() > 0) {
                            for (int i = 0; i < arg.getComputedNSNum(); i++) {
                                ComputedNamespace compNS = (ComputedNamespace) ((EnclosedExpression) arg.getSubExpressions().get(i)).getExpression();
                                evalNeeded = true;
                                compNS.getExpression().accept(this);
                                evalNeeded = false;
                                declarePrefix(compNS.getNcname(), evalResultValue == null ? "" : evalResultValue);
                            }
                            // hasPrefixMapping = true;
                        }
                        // prepare attributes
                        prepareAttributes(arg);
                        // case of attributes (to be evaluated in sons) to add at <start_element>
                        buildAttributes(arg);
                        // case of namespace in top element
                        if (nameSpace != EMPTY_STRING) {
                            String prefix = reconstructionDeclarations.getPrefix(nameSpace);
                            if (prefix == null) {
                                if (declaredDeclarations != null)
                                    prefix = declaredDeclarations.getPrefix(nameSpace);
                                if (prefix != null)
                                    declarePrefix(prefix, nameSpace == null ? "" : nameSpace);
                            }
                        }
                        startPrefixMapping(root);
                        // WRITE START
                        startElement(nameSpace, localName, listAttributes);
                        // set lastSkolemIDsSize
                        this.skolemIDs = arg.getSkolemIDs();
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        // WRITE SONS
                        ArrayList subExpressions = arg.getSubExpressions();
                        if (subExpressions != null) {
                            boolean previousRoot = root;
                            root = false;
                            reconstructionHandler.setPrevIsOnlyChar(false);
                            for (int i = arg.getComputedAttNum() + arg.getComputedNSNum(); i < subExpressions.size(); i++) {
                                XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
                                expr.accept(this);
                                reconstructionHandler.setPrevIsOnlyChar(reconstructionHandler.IsOnlyChars());
                                this.skolemIDs = arg.getSkolemIDs();
                                if (skolemIDs != null)
                                    lastSkolemIDsSize = skolemIDs.size();
                            }
                            root = previousRoot;
                        }
                        // WRITE STOP
                        endElement(nameSpace, localName);
                        if (localName.equals("definition")) {
                            int k = 0;
                            k++;
                        }
                        endPrefixMapping(root);
                        popContext(arg);
                        if (!hasNext()) {
                            break;
                        }
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak) {
                                    break;
                                }
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext()) {
                                break;
                            }
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        pushContext(arg);
                        evalNeeded = true;
                        startTag.accept(this);
                        evalNeeded = false;
                        String nameSpace = evalResultNamespace;
                        String localName = evalResultValue;
                        evalResultNamespace = null;
                        evalResultValue = null;
                        // special for computed namespaces
                        if (arg.getComputedNSNum() > 0) {
                            for (int i = 0; i < arg.getComputedNSNum(); i++) {
                                ComputedNamespace compNS = (ComputedNamespace) ((EnclosedExpression) arg.getSubExpressions().get(i)).getExpression();
                                evalNeeded = true;
                                compNS.getExpression().accept(this);
                                evalNeeded = false;
                                declarePrefix(compNS.getNcname(), evalResultValue == null ? "" : evalResultValue);
                            }
                            // hasPrefixMapping = true;
                        }
                        // prepare attributes
                        prepareAttributes(arg);
                        // case of attributes (to be evaluated in sons) to add at <start_element>
                        buildAttributes(arg);
                        // case of namespace in top element
                        if (nameSpace != EMPTY_STRING) {
                            String prefix = reconstructionDeclarations.getPrefix(nameSpace);
                            if (prefix == null) {
                                if (declaredDeclarations != null)
                                    prefix = declaredDeclarations.getPrefix(nameSpace);
                                if (prefix != null)
                                    declarePrefix(prefix, nameSpace == null ? "" : nameSpace);
                            }
                        }
                        startPrefixMapping(root);
                        startElement(nameSpace, localName, listAttributes);
                        // set lastSkolemIDsSize
                        this.skolemIDs = arg.getSkolemIDs();
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        // WRITE SONS
                        ArrayList subExpressions = arg.getSubExpressions();
                        if (subExpressions != null) {
                            boolean previousRoot = root;
                            root = false;
                            int num = -1;
                            for (int i = arg.getComputedAttNum() + arg.getComputedNSNum(); i < subExpressions.size(); i++) {
                                gbvv.reset();
                                ((XQueryExpression) subExpressions.get(i)).accept(gbvv);
                                List vars = gbvv.getOnlyVariables();
                                if (vars != null && !vars.isEmpty())
                                    // if (!((XQueryExpression) subExpressions.get(i)). instanceof Value))
                                    if (num == -1)
                                        num = i;
                                    else {
                                        num = -1;
                                        break;
                                    }
                            }
                            reconstructionHandler.setPrevIsOnlyChar(false);
                            for (int i = arg.getComputedAttNum() + arg.getComputedNSNum(); i < subExpressions.size(); i++) {
                                XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
                                // CHANGED 10/04/2003
                                if (num == i && arg.getRoot())
                                    expr.setRoot(true);
                                expr.accept(this);
                                reconstructionHandler.setPrevIsOnlyChar(reconstructionHandler.IsOnlyChars());
                                this.skolemIDs = arg.getSkolemIDs();
                                if (skolemIDs != null)
                                    lastSkolemIDsSize = skolemIDs.size();
                            }
                            reconstructionHandler.setIsOnlyChar(false);
                            root = previousRoot;
                        }
                        // WRITE STOP
                        endElement(nameSpace, localName);
                        endPrefixMapping(root);
                        popContext(arg);
                        if (!hasNext()) {
                            break;
                        }
                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext()) {
                                break;
                            }
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot()) {
                            break;
                        }
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        // popContext(arg);
        noReconstructor = false;
    }

    // public void visit(ExternalVariable arg) throws XQueryException {
    // evalVisitor.doGenerate(true);
    // evalVisitor.setReturnType(EvaluationVisitor.NODE_TYPE);
    // arg.accept(evalVisitor);
    // ArrayList list = evalVisitor.getResNodes();
    // evalVisitor.doGenerate(false);
    // }

    private void outputITEExpression(ITEExpression arg) throws XQueryException {
        evalVisitor.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
        arg.getIfExpression().accept(evalVisitor);
        if (evalVisitor.getVerdict())
            arg.getThenExpression().accept(this);
        else
            arg.getElseExpression().accept(this);
        // reset key variables
        skolemIDs = arg.getSkolemIDs();
        if (skolemIDs != null)
            lastSkolemIDsSize = skolemIDs.size();
    }

    public void visit(ITEExpression arg) throws XQueryException { // TODO !!!!! nearly the SAME as Element

        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        skolemIDs = arg.getSkolemIDs();
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    outputITEExpression(arg);
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // case of namespace in top element
                    outputITEExpression(arg);
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                outputITEExpression(arg);
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        outputITEExpression(arg);
                        // WRITE STOP
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    outputITEExpression(arg);
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;

                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        outputITEExpression(arg);
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        outputITEExpression(arg);
                        if (!hasNext())
                            break;

                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        noReconstructor = false;

    }

    public void visit(LibraryFunctionCall arg) throws XQueryException {
        evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
        arg.accept(evalVisitor);
        Comparable comp = evalVisitor.getResValue();
        characters(comp.toString().toCharArray(), 0, comp.toString().length());
    }

    public void visit(EnclosedExpression arg) throws XQueryException {
        reconstructionHandler.setPrevIsOnlyChar(false);
        arg.getExpression().accept(this);
        reconstructionHandler.setIsOnlyChar(false);
    }

    public void visit(ParenthesizedExpression arg) throws XQueryException {
        if (arg.getExpression() == null) {
            // do nothing
            return;
        }
        arg.getExpression().accept(this);
    }

    private void outputLocatedExpression(LocatedExpression arg, int nodeAccessor) throws XQueryException {
        if (!noResultSet) {
            generate(arg.getStringValue(), nodeAccessor, arg.getLoopIDs());
        }
    }

    public void visit(LocatedExpression arg) throws XQueryException { // TODO !!!!! nearly the SAME as Element

        if (arg.isExternal()) {
            evalVisitor.doGenerate(true);
            evalVisitor.setReturnType(EvaluationVisitor.NODE_TYPE);
            arg.accept(evalVisitor);
            evalVisitor.getResNodes();
            evalVisitor.doGenerate(false);
            return;
        }

        if (evalNeeded) {
            evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
            arg.accept(evalVisitor);
            evalResultNamespace = EMPTY_STRING;
            Comparable comp = evalVisitor.getResValue();
            if (comp == null)
                evalResultValue = EMPTY_STRING;
            else {
                if (evalResultValue != null && evalResultValue.length() > 0)
                    evalResultValue += " " + comp.toString();
                else
                    evalResultValue = comp.toString();
            }
            return;
        }

        boolean putPrefixMapping = false;

        // if (hasPrefixMapping) {
        startPrefixMapping(true);
        // hasPrefixMapping = false;
        // putPrefixMapping = true;
        // }

        ArrayList steps = arg.getSteps();
        int nodeAccessor = arg.getNodeAccessor();
        // boolean hasSpecialStep = false;
        if (nodeAccessor == Step.NONE) {
            int lastPosition = steps.size() - 1;
            Step lastStep = (Step) steps.get(lastPosition);
            if (lastStep.getStepKind() % 16 != 0) {
                arg.setNodeAccessor(lastStep.getStepKind());
                nodeAccessor = arg.getNodeAccessor();
                steps.remove(lastPosition);
            }
        }
        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    outputLocatedExpression(arg, nodeAccessor);
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // case of namespace in top element
                    // set lastSkolemIDsSize
                    if (skolemIDs != null)
                        lastSkolemIDsSize = skolemIDs.size();
                    outputLocatedExpression(arg, nodeAccessor);
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                // WRITE SONS
                outputLocatedExpression(arg, nodeAccessor);
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        outputLocatedExpression(arg, nodeAccessor);
                        // WRITE STOP
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    // set lastSkolemIDsSize
                    if (skolemIDs != null)
                        lastSkolemIDsSize = skolemIDs.size();
                    // WRITE SONS
                    outputLocatedExpression(arg, nodeAccessor);
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;

                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        outputLocatedExpression(arg, nodeAccessor);
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        outputLocatedExpression(arg, nodeAccessor);
                        if (!hasNext())
                            break;

                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }

        // if (hasSpecialStep) {
        // steps.add(stepText);
        // }

        noReconstructor = false;
        // if (putPrefixMapping || outsidePutPrefixMapping) {
        // if (putPrefixMapping)
        // hasPrefixMapping = true;
        // if (outsidePutPrefixMapping)
        // outsidePutPrefixMapping = false;
        // putPrefixMapping = false;
        endPrefixMapping(root);
        // }
    }

    public void visit(QName arg) throws XQueryException {
        if (evalNeeded) {
            evalResultNamespace = (arg.getNameSpace() == null) ? EMPTY_STRING : arg.getNameSpace();
            evalResultValue = (arg.getLocalName() == null) ? EMPTY_STRING : arg.getLocalName();
        }
    }

    // -----------------------------------------------------------
    // QNAME FUNCTIONS
    // -----------------------------------------------------------

    public void visit(org.xquark.xquery.parser.primitivefunctions.xsfunctions.FunctionQNAME arg) throws XQueryException {
        if (evalNeeded) {
            arg.getArgument(0).accept(this);
            String prefix = evalResultValue;
            if (prefix == null)
                throw new XQueryException("Could not evaluate " + arg.getArgument(0));
            String namespace = reconstructionDeclarations.getNamespaceURI(prefix);
            if (namespace == null)
                throw new XQueryException("Could not find prefix : " + prefix);
            arg.getArgument(1).accept(this);
            String localname = evalResultValue;
            if (localname == null)
                throw new XQueryException("Could not evaluate " + arg.getArgument(1));
            evalResultNamespace = (namespace == null) ? EMPTY_STRING : namespace;
            evalResultValue = (localname == null) ? EMPTY_STRING : localname;
        }
    }

    public void visit(FunctionQNAME arg) throws XQueryException {
        if (evalNeeded) {
            arg.getArgument(0).accept(this);
            String qname = evalResultValue;
            if (qname == null)
                throw new XQueryException("Could not evaluate " + arg.getArgument(0));
            int index = -1;
            String namespace = null;
            if (qname.startsWith("{")) {
                index = qname.lastIndexOf('}');
                if (index != -1)
                    namespace = qname.substring(1, index);
            } else {
                index = qname.lastIndexOf(":");
                String prefix = null;
                if (index != -1) {
                    prefix = qname.substring(0, index);
                    namespace = reconstructionDeclarations.getNamespaceURI(prefix);
                    if (namespace == null)
                        throw new XQueryException("Could not find prefix : " + prefix);
                }
            }
            String localname = qname.substring(index + 1);
            evalResultNamespace = (namespace == null) ? EMPTY_STRING : namespace;
            evalResultValue = (localname == null) ? EMPTY_STRING : localname;
        }
    }

    // TODO danger when data will be treated in QMEM
    public void visit(FunctionDATA arg) throws XQueryException {
        evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
        arg.accept(evalVisitor);
        if (evalVisitor.getResValue() == null) {
            throw new XQueryException("Could not evaluate : " + arg);
        }
        String res = evalVisitor.getResValue().toString();
        characters(res.toCharArray(), 0, res.length());
    }

    // TODO for now string is buffered here but should be done at lower level for eventual streaming
    private void outputFunctionXML_FRAGMENT(FunctionXML_FRAGMENT arg) throws XQueryException {
        evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
        arg.getArgument(0).accept(evalVisitor);
        if (evalVisitor.getResValue() == null) {
            throw new XQueryException("Could not evaluate : " + arg);
        }
        String res = evalVisitor.getResValue().toString();
        try {
            filter.setQType(arg.getQType());
            filter.parse(new InputSource(new StringReader(res)));
        } catch (SAXException e) {
            throw new XQueryException("Could not parse xml fragment.", e);
        } catch (IOException e) {
            throw new XQueryException("Could not parse xml fragment.", e);
        }
        // reset key variables
        skolemIDs = arg.getSkolemIDs();
        if (skolemIDs != null)
            lastSkolemIDsSize = skolemIDs.size();
    }

    public void visit(FunctionXML_FRAGMENT arg) throws XQueryException { // TODO !!!!! nearly the SAME as Element

        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        skolemIDs = arg.getSkolemIDs();
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    outputFunctionXML_FRAGMENT(arg);
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // case of namespace in top element
                    outputFunctionXML_FRAGMENT(arg);
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                outputFunctionXML_FRAGMENT(arg);
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        outputFunctionXML_FRAGMENT(arg);
                        // WRITE STOP
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    outputFunctionXML_FRAGMENT(arg);
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;

                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        outputFunctionXML_FRAGMENT(arg);
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        outputFunctionXML_FRAGMENT(arg);
                        if (!hasNext())
                            break;

                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        noReconstructor = false;

    }

    public void visit(Value arg) throws XQueryException {
        String text = arg.getValue();
        if (evalNeeded) {
            evalResultNamespace = EMPTY_STRING;
            if (evalResultValue != null && evalResultValue.length() > 0)
                evalResultValue += ' ' + text;
            else
                evalResultValue = text;
        } else {
            characters(text.toCharArray(), 0, text.length());
        }
    }

    private void outputVariable(Variable arg) throws XQueryException {
        if (!noResultSet) {
            int nodeAccessor = Step.NONE;
            if (arg.getExpression() instanceof LocatedExpression) {
                LocatedExpression tmpLoc = (LocatedExpression) arg.getExpression();
                ArrayList steps = tmpLoc.getSteps();
                nodeAccessor = tmpLoc.getNodeAccessor();
                if (nodeAccessor == Step.NONE) {
                    int lastPosition = steps.size() - 1;
                    Step lastStep = (Step) steps.get(lastPosition);
                    if (lastStep.getStepKind() % 16 != 0) {
                        tmpLoc.setNodeAccessor(lastStep.getStepKind());
                        nodeAccessor = tmpLoc.getNodeAccessor();
                        steps.remove(lastPosition);
                    }
                }
                // Step lastStep = (Step) steps.get(steps.size() - 1);
                // XQueryExpression lastStepExpr = lastStep.getExpression();
                // if (lastStepExpr instanceof NodeTest && ((NodeTest) lastStepExpr).getKind() == NodeKind.TEXT)
                // if (lastStep.getAxis() == Axis.CHILD)
                // nodeAccessor = XDBCResultSetInterface.NODE_ACCESSOR_TEXT;
                // else if (lastStep.getAxis() == Axis.SELF)
                // nodeAccessor = XDBCResultSetInterface.NODE_ACCESSOR_SELF_TEXT;
            }
            generate(arg.getStringValue(), nodeAccessor, arg.getLoopIDs());
        }
    }

    public void visit(Variable arg) throws XQueryException { // TODO !!!!! nearly the SAME as Element

        if (arg.isExternal()) {
            if (arg.getQType().isAtom() && arg.getExternalValue() != null) {
                Object obj = arg.getExternalValue();
                if (obj == null)
                    throw new XQueryException("External variable " + arg.toString() + " does not have any value");
                String text = obj.toString();
                if (evalNeeded) {
                    evalResultNamespace = EMPTY_STRING;
                    if (evalResultValue != null && evalResultValue.length() > 0)
                        evalResultValue += ' ' + text;
                    else
                        evalResultValue = text;
                } else {
                    characters(text.toCharArray(), 0, text.length());
                }
                return;
            } else {
                evalVisitor.doGenerate(true);
                // evalVisitor.setReturnType(EvaluationVisitor.NODE_TYPE);
                arg.accept(evalVisitor);
                // ArrayList list = evalVisitor.getResNodes();
                evalVisitor.doGenerate(false);
                return;
            }
        }

        if (evalNeeded) {
            evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
            arg.accept(evalVisitor);
            evalResultNamespace = EMPTY_STRING;
            Comparable comp = evalVisitor.getResValue();
            if (comp == null)
                evalResultValue = EMPTY_STRING;
            else {
                if (evalResultValue != null && evalResultValue.length() > 0)
                    evalResultValue += " " + comp.toString();
                else
                    evalResultValue = comp.toString();
            }
            return;
        }

        boolean putPrefixMapping = false;

        // if (hasPrefixMapping) {
        startPrefixMapping(true);
        // hasPrefixMapping = false;
        // putPrefixMapping = true;
        // }

        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    outputVariable(arg);
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // case of namespace in top element
                    // set lastSkolemIDsSize
                    if (skolemIDs != null)
                        lastSkolemIDsSize = skolemIDs.size();
                    outputVariable(arg);
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                outputVariable(arg);
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        outputVariable(arg);
                        // WRITE STOP
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    // set lastSkolemIDsSize
                    if (skolemIDs != null)
                        lastSkolemIDsSize = skolemIDs.size();
                    // WRITE SONS
                    outputVariable(arg);
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;

                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        outputVariable(arg);
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        outputVariable(arg);
                        if (!hasNext())
                            break;

                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        noReconstructor = false;

        // if (putPrefixMapping || outsidePutPrefixMapping) {
        // if (putPrefixMapping)
        // hasPrefixMapping = true;
        // if (outsidePutPrefixMapping)
        // outsidePutPrefixMapping = false;
        // putPrefixMapping = false;
        endPrefixMapping(root);
        // }
    }

    public void visit(XMLComment arg) throws XQueryException {
        String comment = null;
        evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
        arg.getComment().accept(evalVisitor);
        Comparable comp = evalVisitor.getResValue();
        if (comp != null)
            comment = comp.toString();
        try {
            if (comment != null)
                reconstructionHandler.comment(comment.toCharArray(), 0, comment.length());
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void outputXMLProcessingInstruction(XMLProcessingInstruction arg) throws XQueryException {
        String name = null;
        String value = null;

        XQueryExpression expr = arg.getExpression1();
        evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
        expr.accept(evalVisitor);
        Comparable comp = evalVisitor.getResValue();
        name = (comp == null) ? null : comp.toString();
        expr = arg.getExpression2();
        evalVisitor.setReturnType(EvaluationVisitor.VALUE_TYPE);
        expr.accept(evalVisitor);
        comp = evalVisitor.getResValue();
        value = (comp == null) ? EMPTY_STRING : comp.toString();
        try {
            reconstructionHandler.processingInstruction(name, value);
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    public void visit(XMLProcessingInstruction arg) throws XQueryException {
        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        skolemIDs = arg.getSkolemIDs();
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputXMLProcessingInstruction(arg);
                    root = previousRoot;
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputXMLProcessingInstruction(arg);
                    root = previousRoot;
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                // WRITE SONS
                boolean previousRoot = root;
                root = false;
                outputXMLProcessingInstruction(arg);
                root = previousRoot;
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputXMLProcessingInstruction(arg);
                        root = previousRoot;
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    // WRITE SONS
                    boolean previousRoot = root;
                    root = false;
                    outputXMLProcessingInstruction(arg);
                    root = previousRoot;
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;
                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputXMLProcessingInstruction(arg);
                        root = previousRoot;
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        // WRITE SONS
                        boolean previousRoot = root;
                        root = false;
                        outputXMLProcessingInstruction(arg);
                        root = previousRoot;
                        if (!hasNext())
                            break;
                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        noReconstructor = false;
    }

    public void visit(XQueryExpression arg) throws XQueryException {
        throw new XQueryException("Class " + arg.getClass().getName() + " not supported in ReconstructionVisitor.");
    }

    private void writeSubExpressions(XQueryExpression arg, ArrayList subExpressions, boolean special) throws XQueryException {
        if (subExpressions == null)
            return;
        boolean previousRoot = root;
        byte num = 0;
        if (special)
            for (int i = 0; i < subExpressions.size() && num < 2; i++) {
                if (!((XQueryExpression) subExpressions.get(i) instanceof Value))
                    num++;
            }
        reconstructionHandler.setPrevIsOnlyChar(false);
        for (int i = 0; i < subExpressions.size(); i++) {
            XQueryExpression expri = (XQueryExpression) subExpressions.get(i);
            if (special && num == 1 && arg.getRoot())
                expri.setRoot(true);
            expri.accept(this);
            reconstructionHandler.setPrevIsOnlyChar(reconstructionHandler.IsOnlyChars());
            skolemIDs = arg.getSkolemIDs();
            if (skolemIDs != null)
                lastSkolemIDsSize = skolemIDs.size();
        }
        reconstructionHandler.setIsOnlyChar(false);
        root = previousRoot;
    }

    public void visit(XQueryExpressionSequence arg) throws XQueryException { // TODO !!!!! nearly the SAME as Element

        boolean isSkolemID = false;
        boolean isDependID = false;
        noReconstructor = true;
        skolemIDs = arg.getSkolemIDs();
        if ((skolemIDs != null && !skolemIDs.isEmpty()) || arg.getRoot())
            isSkolemID = true;
        ArrayList dependIDs = arg.getDependIDs();
        if (dependIDs != null && !dependIDs.isEmpty())
            isDependID = true;
        boolean loop = arg.getLoop();

        // CASE : this element has no skolem ids
        if (!isSkolemID) {
            // CASE : the element is written until there are no more results available
            if (loop) {
                // no not continue if dependIDs changed
                if (this.noResultSet && dependIDs != lastDependIDs)
                    return;
                while (true) {
                    // WRITE SONS
                    writeSubExpressions(arg, arg.getSubExpressions(), false);
                    noReconstructor = true;
                    if (hasNext())
                        nextAsSAX();
                    else
                        break;
                }
            }
            // CASE : no loop and element has depend ids
            else if (isDependID) {
                boolean noIdentifier = false;
                boolean changeDependIDs = true;
                // / !!! DANGER !!! THIS is stupid code!!! ( )
                try {
                    this.setIdentifiers(false);
                } catch (XQueryException e) {
                    changeDependIDs = false;
                    noIdentifier = true;
                }
                while (noIdentifier || (!isEmpty(dependIDs) && changeDependIDs)) {
                    // case of namespace in top element
                    // set lastSkolemIDsSize
                    if (skolemIDs != null)
                        lastSkolemIDsSize = skolemIDs.size();
                    // WRITE SONS
                    writeSubExpressions(arg, arg.getSubExpressions(), false);
                    if (!hasNext())
                        break;
                    if (!noIdentifier)
                        changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    else
                        changeDependIDs = true;
                    if (!changeDependIDs) {
                        noNext = false;
                        break;
                    } else if (!isEmpty(dependIDs))
                        noNext = false;
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!noIdentifier)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        if (!noIdentifier)
                            this.setIdentifiers(true);
                    }
                    if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                        noNext = true;
                }
            } else {
                // WRITE SONS
                writeSubExpressions(arg, arg.getSubExpressions(), false);
            }
        }
        // there is at least a skolemID
        else {

            // set the current identifier and the next
            if (isSkolemID && !arg.getRoot()) {
                this.setIdentifiers(false);
            }
            // there is no dependID
            if (dependIDs == null && !arg.getRoot()) {
                // CODE NOT USED !!! AS FAR AS I CAN TELL
                if (loop) {
                    while (true) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        // WRITE SONS
                        writeSubExpressions(arg, arg.getSubExpressions(), false);
                        // WRITE STOP
                        noReconstructor = true;
                        noNext = false;
                        if (hasNext())
                            nextAsSAX();
                        else
                            break;
                    }
                } else {
                    // set lastSkolemIDsSize
                    if (skolemIDs != null)
                        lastSkolemIDsSize = skolemIDs.size();
                    // WRITE SONS
                    writeSubExpressions(arg, arg.getSubExpressions(), false);
                }
            } else // dependIDs != null or arg element is root
            {
                noNext = false;
                if (!arg.getRoot() && lastSkolemIDsSize < skolemIDs.size()) {
                    boolean changeDependIDs = true;

                    boolean changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                    // stop if the skolemIDs changed, and that dependIDs does not have an incidence
                    if (dependIDs != null && dependIDs.isEmpty() && changeSkolemIDs) {
                        return;
                    }

                    int firstDependIDPosition = 0;
                    if (dependIDs != null && !dependIDs.isEmpty()) {
                        firstDependIDPosition = getPosition(dependIDs.get(0)) + 1;
                    }
                    int lastSkolemIDPosition = getPosition(skolemIDs.get(skolemIDs.size() - 1)) + 1;
                    int validSkolemNumber = skolemIDs.size();
                    if (skolemIDs.size() > 0) {
                        int index = skolemIDs.indexOf(dependIDs.get(0));
                        if (index != -1)
                            validSkolemNumber = index;
                    }

                    int skipEntries = -1;
                    if (firstDependIDPosition > lastSkolemIDPosition) {
                        skipEntries = validSkolemNumber;
                        while (!changeSkolemIDs && isEmpty(dependIDs)) {
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, lastSkolemIDsSize);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, lastSkolemIDsSize);
                        }
                    }
                    while (!isEmpty(dependIDs) && changeDependIDs) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        // WRITE SONS
                        writeSubExpressions(arg, arg.getSubExpressions(), false);
                        if (!hasNext())
                            break;
                        if (skipEntries == -1) {
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeDependIDs) {
                                noNext = false;
                                break;
                            } else if (!isEmpty(dependIDs))
                                noNext = false;
                        }
                        // ----------------------------
                        // ADDED
                        // -----------------------------
                        else {
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                            changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, skolemIDs, dependIDs, validSkolemNumber);
                            if (!changeSkolemIDs) {
                                noNext = false;
                                boolean noDependIDs = true;
                                boolean makeBreak = false;
                                noReInit = false;
                                while (!changeSkolemIDs && noDependIDs) {
                                    noReconstructor = true;
                                    if (!hasNext()) {
                                        makeBreak = true;
                                        break;
                                    }
                                    identifierCopy(currentIdentifiers, lastIdentifiers);
                                    nextAsSAX();
                                    this.setIdentifiers(true);
                                    changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, validSkolemNumber);
                                    noDependIDs = isEmpty(dependIDs);
                                }
                                if (makeBreak)
                                    break;
                                noNext = true;
                                if (!noDependIDs)
                                    noReInit = true;
                                changeDependIDs = true;
                            } else {
                                if (!isEmpty(dependIDs)) {
                                    noNext = false;
                                    // ADDED 01/04/05
                                    changeDependIDs = false;
                                }
                            }
                        }
                        // ----------------------------
                        // END ADDED
                        // -----------------------------
                        if (!noNext && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (skipEntries == -1) {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs))
                                noNext = true;
                        } else {
                            if (changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs, skipEntries))
                                noNext = false; // CHANGE 03/07/2002 was initially true
                        }
                    }
                } else {
                    boolean changeSkolemIDs = false;
                    while (arg.getRoot() || (!isEmpty(dependIDs) && !changeSkolemIDs)) {
                        // set lastSkolemIDsSize
                        if (skolemIDs != null)
                            lastSkolemIDsSize = skolemIDs.size();
                        // set last dependIDs
                        lastDependIDs = dependIDs;
                        // WRITE SONS
                        writeSubExpressions(arg, arg.getSubExpressions(), true);
                        if (!hasNext())
                            break;

                        if (!arg.getRoot())
                            changeSkolemIDs = changeSkolemIDs(currentIdentifiers, nextIdentifiers, skolemIDs);
                        if (!noNext && !arg.getRoot() && !isEmpty(dependIDs) && !changeSkolemIDs) {
                            noReconstructor = true;
                            if (!hasNext())
                                break;
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                            nextAsSAX();
                            this.setIdentifiers(true);
                        }
                        if (changeSkolemIDs)
                            noNext = true;
                        if (arg.getRoot())
                            break;
                    }
                    // while (!changeSkolemIDs);
                }
                // noNext = true;
            }
        }
        noReconstructor = false;
    }

    /*
     * returns -1 if error 0 if no change n for change in order n element
     */
    // private int identifierCompare(Object[] list1, Object[] list2) throws ReconstructionException {
    // if (list1 == null && list2 != null) {
    // return list2.length;
    // }
    // if (list1 == null && list2 == null) {
    // throw new ReconstructionException("Incorrects argumments in method identifierCompare");
    // }
    // int len1 = list1.length;
    // // if (len1 > skolemSize) len1 = skolemSize;
    // int len2 = list2.length;
    // // if (len2 > skolemSize) len2 = skolemSize;
    // int i = 0;
    // while (i < len1 && i < len2) {
    // if (list1[i] == null || list2[i] == null) {
    // return i + 1;
    // }
    // if (!list1[i].equals(list2[i])) {
    // return i + 1;
    // }
    // i++;
    // }
    // if (len1 != len2) {
    // return i + 1;
    // }
    // return 0;
    // }
    private boolean changeDependIDs(Object[] list1, Object[] list2, ArrayList dependIDs) throws ReconstructionException {
        return changeDependIDs(list1, list2, null, dependIDs, -1);
    }

    private boolean changeDependIDs(Object[] list1, Object[] list2, ArrayList skolemIDs, ArrayList dependIDs, int maxSkolemSize) throws ReconstructionException {
        if (list1 == null && list2 != null) {
            return true;
        }
        if (list1 == null && list2 == null) {
            return true;
        }

        if (skolemIDs != null) {
            int position = -1;
            for (int i = 0; i < maxSkolemSize; i++) {
                position = getPosition((Variable) skolemIDs.get(i));
                // if (list1[position] == null && list2[position] == null) continue; // 2002/09/23
                /*
                 * if (list1[position] == null || list2[position] == null) return true;
                 */
                if (list1[position] == null && list2[position] != null)
                    return false;
                if (!list1[position].equals(list2[position])) {
                    return false;
                }
            }
        }

        for (int j = 0; j < dependIDs.size(); j++) {
            int position = getPosition((Variable) dependIDs.get(j));
            if (list1[position] == null || list2[position] == null)
                return true;
            if (!list1[position].equals(list2[position])) {
                return true;
            }
        }
        return false;
    }

    private boolean changeSkolemIDs(Object[] list1, Object[] list2, ArrayList skolemIDs) throws ReconstructionException {
        return changeSkolemIDs(list1, list2, skolemIDs, (skolemIDs == null) ? 0 : skolemIDs.size());
    }

    private boolean changeSkolemIDs(Object[] list1, Object[] list2, ArrayList skolemIDs, int maxSkolemSize) throws ReconstructionException {
        int position = -1;
        for (int j = 0; j < maxSkolemSize; j++) {
            position = getPosition((Variable) skolemIDs.get(j));
            if (list1[position] == null || list2[position] == null)
                return true;
            if (!list1[position].equals(list2[position])) {
                return true;
            }
        }
        return false;
    }

    private void identifierCopy(Object[] original, Object[] copy) {
        if (copy == null)
            copy = new Object[size];
        for (int i = 0; i < size; i++)
            copy[i] = original[i];
    }

    private boolean isEmpty(ArrayList dependIDs) throws ReconstructionException {
        if (dependIDs == null)
            return true;
        int countIsEmpty = 0;
        for (int i = 0; i < dependIDs.size(); i++) {
            Object identifier = currentIdentifiers[getPosition((Variable) dependIDs.get(i))];
            if (identifier == null || (identifier instanceof ArrayList && ((ArrayList) identifier).size() == 1 && ((ArrayList) identifier).get(0) == null))
                countIsEmpty++;
        }
        if (countIsEmpty == dependIDs.size()) {
            return true;
        }
        return false;
    }

    private int getPosition(Object obj) throws ReconstructionException {
        int position = varList.indexOf(obj);
        if (position == -1)
            throw new ReconstructionException("This variable " + obj + " is not defined in the list of variables.");
        return position;
    }

    private String fetch(String path, int nodeAccessor, String loopid) throws XQueryException {
        try {
            return xdbcResultSet.fetch(path, nodeAccessor, loopid, psvisp);
        } catch (XMLDBCException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void generate(String path, int nodeAccessor, String loopid) throws XQueryException {
        // if (nodeAccessor == XDBCResultSetInterface.NODE_ACCESSOR_SELF_TEXT || nodeAccessor ==
        // XDBCResultSetInterface.NODE_ACCESSOR_TEXT)
        // hasCharacters = true;
        // else
        // hasCharacters = false;
        try {
            xdbcResultSet.generate(path, nodeAccessor, loopid, psvisp);
        } catch (XMLDBCException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void characters(char[] ch, int start, int length) throws XQueryException {
        try {
            reconstructionHandler.characters(ch, start, length);
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void startElement(String nameSpace, String localName, AttributesImpl atts) throws XQueryException {
        try {
            Collection prefixes = reconstructionDeclarations.getPrefixes(nameSpace);
            if (prefixes.isEmpty())
                reconstructionHandler.startElement(nameSpace, localName, localName, atts);
            else
                reconstructionHandler.startElement(nameSpace, localName, prefixes.iterator().next() + ":" + localName, atts);
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void endElement(String nameSpace, String localName) throws XQueryException {
        try {
            Collection prefixes = reconstructionDeclarations.getPrefixes(nameSpace);
            if (prefixes.isEmpty())
                reconstructionHandler.endElement(nameSpace, localName, localName);
            else
                reconstructionHandler.endElement(nameSpace, localName, prefixes.iterator().next() + ":" + localName);
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private boolean hasNext() throws XQueryException {
        try {
            return resultSet.hasNext();
        } catch (XMLDBCException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void nextAsSAX() throws XQueryException {
        try {
            resultSet.nextAsSAX();
        } catch (XMLDBCException e) {
            throw new XQueryException(e.getMessage());
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void startPrefixMapping(boolean first) throws XQueryException {
        try {
            String prefix = null;
            ArrayList currentList = null;
            if (first)
                currentList = new ArrayList(reconstructionDeclarations.getPrefixes());
            else
                currentList = new ArrayList(reconstructionDeclarations.getDeclaredPrefixes());
            for (int i = 0; i < currentList.size(); i++) {
                prefix = (String) currentList.get(i);
                reconstructionHandler.startPrefixMapping(prefix, reconstructionDeclarations.getNamespaceURI(prefix));
            }
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void endPrefixMapping(boolean first) throws XQueryException {
        try {
            ArrayList currentList = null;
            if (first)
                currentList = new ArrayList(reconstructionDeclarations.getPrefixes());
            else
                currentList = new ArrayList(reconstructionDeclarations.getDeclaredPrefixes());
            for (int i = 0; i < currentList.size(); i++) {
                reconstructionHandler.endPrefixMapping((String) currentList.get(i));
            }
        } catch (SAXException e) {
            throw new XQueryException(e.getMessage());
        }
    }

    private void prepareAttributes(Element elt) throws XQueryException {
        listAttributes.clear();
        ArrayList attributes = elt.getAttributes();
        if (attributes != null) {
            for (int i = 0; i < attributes.size(); i++) {
                AttributeValuePair atti = (AttributeValuePair) attributes.get(i);
                if (atti.isXmlns()) {
                    if (atti.getAttributeValue() instanceof ValueString)
                        declarePrefix(((ValueString) atti.getAttributeName()).getStringValue(), ((ValueString) atti.getAttributeValue()).getStringValue());
                    else
                        atti.accept(this);
                } else
                    atti.accept(this);
            }
        }
    }

    // TODO ANNUAIRE the whole function should be revievew
    private void buildAttributes(Element arg) throws XQueryException {
        ArrayList sons = arg.getSubExpressions();
        if (sons == null)
            return;
        for (int i = arg.getComputedNSNum(); i < arg.getComputedNSNum() + arg.getComputedAttNum(); i++) {
            XQueryExpression son = (XQueryExpression) sons.get(i);
            // change LARS 01/07/04 this problem is catched later, avoid throwing exception if typing is lax
            // if (son.getQType().isMultiple()) {
            // throw new ReconstructionException("An element cannot have several attributes with the same name.");
            // }
            ArrayList dependIDs = son.getDependIDs();
            boolean root = son.getRoot();

            while (true) {
                if (son instanceof EnclosedExpression)
                    son = ((EnclosedExpression) son).getExpression();
                else if (son instanceof ParenthesizedExpression)
                    son = ((ParenthesizedExpression) son).getExpression();
                else if (son instanceof XQueryExpressionSequence) {
                    if (((XQueryExpressionSequence) son).getSubExpressions().size() == 1)
                        son = (XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0);
                    else
                        continue;
                } else
                    break;
            }

            if (dependIDs == null || dependIDs.isEmpty()) {
                if (sons.size() == 1 && arg.getRoot())
                    son.setRoot(true);
                // evalNeeded = true;
                // son.accept(this);
                // evalNeeded = false;
                QType qtype = son.getQType();
                String value = null;
                // TODO ANNUAIRE this is dangerous if attribute in then and something else not empty in else
                if (son instanceof ITEExpression) {
                    evalNeeded = true;
                    son.accept(this);
                    evalNeeded = false;
                } else {
                    String nameSpaceAtt = ((QName) qtype.getName()).getNameSpace();
                    String localNameAtt = ((QName) qtype.getName()).getLocalName();
                    if (son instanceof Variable) {
                        if (!noResultSet)
                            value = fetch(son.getStringValue(), 0, son.getLoopIDs());
                    } else if (son instanceof LocatedExpression || son instanceof AttributeValuePair) { // || son instanceof EnclosedExpression || son instanceof ParenthesizedExpression) {
                        evalNeeded = true;
                        son.accept(this);
                        evalNeeded = false;
                        value = evalResultValue;
                        // } else if (son instanceof XQueryExpressionSequence && ((XQueryExpressionSequence) son).getSubExpressions().size() == 1) {
                        // evalNeeded = true;
                        // XQueryExpression firstSon = (XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0);
                        // firstSon.accept(this);
                        // evalNeeded = false;
                        // // // DANGER!!! Olivier: 2002/12/17
                        // // if (firstSon instanceof AttributeValuePair && ((AttributeValuePair) firstSon).isComputed())
                        // // continue;
                        // value = evalResultValue;
                        // TODO ANNUAIRE this is dangerous if attribute in then and something else not empty in else
                    } else if (son instanceof ITEExpression) {
                    } else
                        break;
                    if (!(son instanceof AttributeValuePair)) {
                        // test if attribute already exists
                        // erase 29/06/04 put again 06/08/07
                        if (listAttributes.getValue((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt) != null)
                            throw new ReconstructionException("An element cannot have several attributes with the same name.");
                        String prefix = null;
                        if (nameSpaceAtt != EMPTY_STRING) {
                            prefix = reconstructionDeclarations.getPrefix(nameSpaceAtt);
                            if (prefix == null) {
                                if (declaredDeclarations != null)
                                    prefix = declaredDeclarations.getPrefix(nameSpaceAtt);
                                if (prefix != null)
                                    declarePrefix(prefix, nameSpaceAtt == null ? "" : nameSpaceAtt);
                            }
                        }
                        // generate attribute
                        if (listAttributes.getValue((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt) == null)
                            listAttributes.addAttribute((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt, (prefix == null ? "" : prefix + ":") + localNameAtt, "STRING", (value == null ? "" : value));
                    } else {
                        // TODO ANNUAIRE this is only temporary should be enhhanced
                        break;
                    }
                }
            } else {
                if (son instanceof XQueryExpressionSequence) {
                    dependIDs = ((XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0)).getDependIDs();
                    root = ((XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0)).getRoot();
                }
                noNext = false;
                boolean changeDependIDs = true;
                this.setIdentifiers(false);
                QType qtype = son.getQType();
                String value = null;
                String nameSpaceAtt = null;
                String localNameAtt = null;

                while (!isEmpty(dependIDs) && changeDependIDs) {

                    // TODO ANNUAIRE this is dangerous if attribute in then and something else not empty in else
                    if (son instanceof ITEExpression) {
                        evalNeeded = true;
                        son.accept(this);
                        evalNeeded = false;
                    } else {
                        // get nameSpaceAtt, localNameAtt and value from expression
                        nameSpaceAtt = ((QName) qtype.getName()).getNameSpace();
                        localNameAtt = ((QName) qtype.getName()).getLocalName();
                        if (son instanceof Variable) {
                            if (!noResultSet)
                                value = fetch(son.getStringValue(), 0, son.getLoopIDs());
                        } else if (son instanceof LocatedExpression || son instanceof AttributeValuePair) { // || son instanceof EnclosedExpression || son instanceof ParenthesizedExpression) {
                            evalNeeded = true;
                            son.accept(this);
                            evalNeeded = false;
                            value = evalResultValue;
                            // } else if (son instanceof XQueryExpressionSequence && ((XQueryExpressionSequence) son).getSubExpressions().size() == 1) {
                            // evalNeeded = true;
                            // ((XQueryExpression) ((XQueryExpressionSequence) son).getSubExpressions().get(0)).accept(this);
                            // evalNeeded = false;
                            // value = evalResultValue;
                        } else
                            break;
                        if (!(son instanceof AttributeValuePair)) {
                            // test if attribute already exists
                            if (listAttributes.getValue((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt) != null)
                                throw new ReconstructionException("An element cannot have several attributes with the same name.");
                            String prefix = null;
                            if (nameSpaceAtt != EMPTY_STRING) {
                                prefix = reconstructionDeclarations.getPrefix(nameSpaceAtt);
                                if (prefix == null) {
                                    if (declaredDeclarations != null)
                                        prefix = declaredDeclarations.getPrefix(nameSpaceAtt);
                                    if (prefix != null)
                                        declarePrefix(prefix, nameSpaceAtt == null ? "" : nameSpaceAtt);
                                }
                            }
                            // generate attribute
                            listAttributes.addAttribute((nameSpaceAtt == null ? "" : nameSpaceAtt), localNameAtt, (prefix == null ? "" : prefix + ":") + localNameAtt, "STRING", (value == null ? "" : value));
                        } else {
                            // TODO ANNUAIRE this is only temporary should be enhhanced
                            break;
                        }
                    }
                    this.setIdentifiers(false);
                    if (!hasNext())
                        break;
                    changeDependIDs = changeDependIDs(currentIdentifiers, nextIdentifiers, dependIDs);
                    if (!changeDependIDs) {
                        noNext = true;
                        break;
                    }
                    if (!noNext) {
                        noReconstructor = true;
                        if (!hasNext())
                            break;
                        if (!root)
                            identifierCopy(currentIdentifiers, lastIdentifiers);
                        nextAsSAX();
                        this.setIdentifiers(true);
                        changeDependIDs = changeDependIDs(lastIdentifiers, currentIdentifiers, dependIDs);
                    }
                }
            }
        }
    }

    private void popContext(Element arg) {
        // System.out.println("popContext " + arg.getStartTag());
        reconstructionDeclarations.popContext();
    }

    private void pushContext(Element arg) {
        // System.out.println("pushContext " + arg.getStartTag());
        reconstructionDeclarations.pushContext();
    }

    private void declarePrefix(String prefix, String uri) throws ReconstructionException {
        reconstructionDeclarations.declarePrefix(prefix, uri);
    }
}
