/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import java.util.List;

import org.xquark.schema.SimpleType;
import org.xquark.schema.Type;
import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.schema.datatypes.PrimitiveUnionType;
import org.xquark.xquery.parser.XQueryExpression;

public abstract class QType implements QTypeVisitable, Cloneable {
    private static final String RCSRevision = "$Revision: 1.16 $";

    private static final String RCSName = "$Name:  $";

    // occurence types
    public static final byte OCC_0_0 = 0;

    public static final byte OCC_0_1 = 1;

    public static final byte OCC_1_1 = 2;

    public static final byte OCC_0_N = 3;

    public static final byte OCC_1_N = 4;

    // Qtype types
    public static final byte ELEMENT = 1;

    public static final byte ATOM = 2;

    public static final byte ATTRIBUTE = 3;

    public static final byte UNION = 4;

    public static final byte SEQUENCE = 5;

    public static final byte COMMENT = 6;

    public static final byte PROCESSING = 7;

    public static final byte CDATA = 8;

    public static final byte DOCUMENT = 9;

    public static final byte TEXT = 10;

    // primitive XQuery Types

    public static final byte PRIMITIVE_anySimpleType = 0;

    public static final byte PRIMITIVE_string = 1;

    public static final byte PRIMITIVE_float = 2;

    public static final byte PRIMITIVE_double = 3;

    public static final byte PRIMITIVE_decimal = 4;

    public static final byte PRIMITIVE_integer = 5;

    public static final byte PRIMITIVE_duration = 6;

    public static final byte PRIMITIVE_yearMonthDuration = 7;

    public static final byte PRIMITIVE_dayTimeDuration = 8;

    public static final byte PRIMITIVE_dateTime = 9;

    public static final byte PRIMITIVE_time = 10;

    public static final byte PRIMITIVE_date = 11;

    public static final byte PRIMITIVE_gYearMonth = 12;

    public static final byte PRIMITIVE_gYear = 13;

    public static final byte PRIMITIVE_gMonthDay = 14;

    public static final byte PRIMITIVE_gDay = 15;

    public static final byte PRIMITIVE_gMonth = 16;

    public static final byte PRIMITIVE_boolean = 17;

    public static final byte PRIMITIVE_base64Binary = 18;

    public static final byte PRIMITIVE_hexBinary = 19;

    public static final byte PRIMITIVE_anyURI = 20;

    public static final byte PRIMITIVE_Qname = 21;

    public static final byte PRIMITIVE_NOTATION = 22;

    /*
     * 
     * aST = anySimpleType aURI = anyURI b64 = base64Binary bool = boolean dat = date Day = gDay dbl = double dec = decimal dT = dateTime dTD = dayTimeDuration dur = duration flt = float hxB = hexBinary MD = gMonthDay Mon = gMonth NOT = NOTATION QN = Qname str = string tim = time YM = gYearMonth yMD = yearMonthDuration Yr = gYear
     * 
     * S\T aST str flt dbl dec int dur yMD dTD  dT tim dat  YM  Yr  MD Day Mon bool b64 hxB aURI QN NOT 
     * aST  Y   Y   M   M   M   M   M   M   M   M   M   M   M   M   M   M   M   M    M   M   M   N   N 
     * str  Y   Y   M   M   M   M   M   M   M   M   M   M   M   M   M   M   M   M    M   M   M   M   M 
     * flt  Y   Y   Y   Y   M   M   N   N   N   N   N   N   N   N   N   N   N   Y    N   N   N   N   N 
     * dbl  Y   Y   Y   Y   M   M   N   N   N   N   N   N   N   N   N   N   N   Y    N   N   N   N   N 
     * dec  Y   Y   Y   Y   Y   Y   N   N   N   N   N   N   N   N   N   N   N   Y    N   N   N   N   N 
     * int  Y   Y   Y   Y   Y   Y   N   N   N   N   N   N   N   N   N   N   N   Y    N   N   N   N   N 
     * dur  Y   Y   N   N   N   N   Y   Y   Y   N   N   N   N   N   N   N   N   N    N   N   N   N   N
     * yMD  Y   Y   N   N   N   N   Y   Y   Y   N   N   N   N   N   N   N   N   N    N   N   N   N   N
     * dTD  Y   Y   N   N   N   N   Y   Y   Y   N   N   N   N   N   N   N   N   N    N   N   N   N   N
     * dT   Y   Y   N   N   N   N   N   N   N   Y   Y   Y   Y   Y   Y   Y   Y   N    N   N   N   N   N
     * tim  Y   Y   N   N   N   N   N   N   N   N   Y   N   N   N   N   N   N   N    N   N   N   N   N
     * dat  Y   Y   N   N   N   N   N   N   N   Y   N   Y   Y   Y   Y   Y   Y   N    N   N   N   N   N 
     * YM   Y   Y   N   N   N   N   N   N   N   N   N   N   Y   N   N   N   N   N    N   N   N   N   N
     * Yr   Y   Y   N   N   N   N   N   N   N   N   N   N   N   Y   N   N   N   N    N   N   N   N   N
     * MD   Y   Y   N   N   N   N   N   N   N   N   N   N   N   N   Y   N   N   N    N   N   N   N   N
     * Day  Y   Y   N   N   N   N   N   N   N   N   N   N   N   N   N   Y   N   N    N   N   N   N   N
     * Mon  Y   Y   N   N   N   N   N   N   N   N   N   N   N   N   N   N   Y   N    N   N   N   N   N
     * bool Y   Y   Y   Y   Y   Y   N   N   N   N   N   N   N   N   N   N   N   Y    N   N   N   N   N 
     * b64  Y   Y   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N    Y   Y   N   N   N
     * hxB  Y   Y   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N    Y   Y   N   N   N
     * aURI Y   Y   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N    N   N   Y   N   N
     * QN   Y   Y   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N    N   N   N   Y   N
     * NOT  Y   Y   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N   N    N   N   N   N   M
     */
    // NOTE : Maybe is Yes with restrictions!
    private static final boolean castTab[][] = { {
    // aST str flt dbl dec int dur yMD dTD dT tim dat YM Yr MD Day Mon bool b64 hxB aURI QN NOT
    /* aST */
    true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true, false, false }, {
    /* str */
    true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true,  true }, {
    /* flt */
    true,  true,  true,  true,  true,  true, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false }, {
    /* dbl */
    true,  true,  true,  true,  true,  true, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false }, {
    /* dec */
    true,  true,  true,  true,  true,  true, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false }, {
    /* int */
    true,  true,  true,  true,  true,  true, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false }, {
    /* dur */
    true,  true, false, false, false, false,  true,  true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false }, {
    /* yMD */
    true,  true, false, false, false, false,  true,  true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false }, {
    /* dTD */
    true,  true, false, false, false, false,  true,  true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false }, {
    /* dT */
    true,  true, false, false, false, false, false, false, false,  true,  true,  true,  true,  true,  true,  true,  true, false, false, false, false, false, false }, {
    /* tim */
    true,  true, false, false, false, false, false, false, false, false,  true, false, false, false, false, false, false, false, false, false, false, false, false }, {
    /* dat */
    true,  true, false, false, false, false, false, false, false,  true, false,  true,  true,  true,  true,  true,  true, false, false, false, false, false, false }, {
    /* YM */
    true,  true, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false, false, false, false, false, false }, {
    /* Yr */
    true,  true, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false, false, false, false, false }, {
    /* MD */
    true,  true, false, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false, false, false, false }, {
    /* Day */
    true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false, false, false }, {
    /* Mon */
    true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false, false }, {
    /* bool */
    true,  true,  true,  true,  true,  true, false, false, false, false, false, false, false, false, false, false, false,  true, false, false, false, false, false }, {
    /* b64 */
    true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,  true,  true, false, false, false }, {
    /* hxB */
    true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false , true,  true, false, false, false }, {
    /* aURI */
    true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,  true, false, false }, {
    /* QN */
    true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,  true, false }, {
    /* NOT */
    true,  true, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false, false,  true } };

    protected byte subclass = 0;

    protected byte occurence = OCC_1_1;

    // protected int min = 1;
    // protected int max = 1;
    protected boolean nillable = true;

    public static byte computeOccurence(byte occ1, byte occ2) {
        switch (occ1) {
        case OCC_1_1:
            return occ2;
        case OCC_0_1:
            switch (occ2) {
            case OCC_1_1:
            case OCC_0_1:
                return OCC_0_1;
            case OCC_1_N:
            case OCC_0_N:
                return OCC_0_N;
            }
        case OCC_1_N:
            switch (occ2) {
            case OCC_1_1:
            case OCC_1_N:
                return OCC_1_N;
            case OCC_0_1:
            case OCC_0_N:
                return OCC_0_N;
            }
        default:
            return occ1;
        }
    }

    public static byte getOccurence(int min, int max) {
        if (min > 0) {
            if (max > 1)
                return QType.OCC_1_N;
            else
                return QType.OCC_1_1;
        } else {
            if (max > 1)
                return QType.OCC_0_N;
            else
                return QType.OCC_0_1;
        }
    }

    // methods on occurence
    public byte getOccurence() {
        return occurence;
    }

    public boolean isNillable() {
        return nillable;
    }

    public QType changeOccurence(byte occurence) {
        QType qtype = null;
        try {
            qtype = (QType) clone();
        } catch (CloneNotSupportedException cnse) {
        }
        qtype.occurence = occurence;
        return qtype;
    }

    public int getMaxOccurence() {
        switch (occurence) {
        case OCC_0_0:
            return 0;
        case OCC_1_1:
        case OCC_0_1:
            return 1;
        case OCC_0_N:
        case OCC_1_N:
            return Integer.MAX_VALUE;
        }
        return -1;
    }

    public int getMinOccurence() {
        switch (occurence) {
        case OCC_0_N:
        case OCC_0_1:
        case OCC_0_0:
            return 0;
        case OCC_1_1:
        case OCC_1_N:
            return 1;
        }
        return -1;
    }

    public byte getSubClass() {
        return subclass;
    }

    public SimpleType getSimpleType() {
        return null;
    }

    public Type getType() {
        return null;
    }

    public List getList() {
        return null;
    }

    public XQueryExpression getName() {
        return null;
    }

    abstract public Object clone() throws CloneNotSupportedException;

    // tells whether qtype is numeric ( DECIMAL, DOUBLE, FLOAT )
    // QTypeAtom with type numeric
    // QTypeAttribute with type numeric
    // QTypeElement with type numeric
    // QTypeCollection with type verifying isNumeric
    // QTypeSequence with all types verifying isNumeric
    // QTypeUnion with at leat one type verifying isNumeric
    public boolean isNumeric() {
        return false;
    }

    // tells whether qtype is boolean
    // QTypeAtom with type boolean
    // QTypeAttribute with type boolean
    // QTypeElement with type boolean
    // QTypeCollection with type verifying isBoolean
    // QTypeSequence with all types verifying isBoolean
    // QTypeUnion with at leat one type verifying isBoolean
    public boolean isBoolean() {
        return false;
    }

    // tells whether qtype is integer
    // QTypeAtom with type integer
    // QTypeAttribute with type integer
    // QTypeElement with type integer
    // QTypeCollection with type verifying isInteger
    // QTypeSequence with all types verifying isInteger
    // QTypeUnion with at leat one type verifying isInteger
    public boolean isInteger() {
        return false;
    }

    // tells whether qtype is string
    // QTypeAtom with type string
    // QTypeAttribute with type string
    // QTypeElement with type string
    // QTypeCollection with type verifying isString
    // QTypeSequence with all types verifying isString
    // QTypeUnion with at leat one type verifying isString
    public boolean isString() {
        return false;
    }

    // tells whether qtype is an atom
    // QTypeAtom
    // QTypeCollection with type verifying isAtom
    // QTypeSequence with all types verifying isAtom
    // QTypeUnion with at leat one type verifying isAtom
    public boolean isAtom() {
        return false;
    }

    // tells whether qtype can react to text()
    // QTypeAttribute
    // QTypeElement
    // QTypeCollection with type verifying isText
    // QTypeSequence with all types verifying isText
    // QTypeUnion with at leat one type verifying isText
    public boolean isText() {
        return false;
    }

    // tells whether qtype is a date
    // QTypeAtom with type date
    // QTypeAttribute with type date
    // QTypeElement with type date
    // QTypeCollection with type verifying isDate
    // QTypeSequence with all types verifying isDate
    // QTypeUnion with at leat one type verifying isDate
    public boolean isDate() {
        return false;
    }

    // tells whether qtype is a dateTime
    // QTypeAtom with type dateTime
    // QTypeAttribute with type dateTime
    // QTypeElement with type dateTime
    // QTypeCollection with type verifying isDateTime
    // QTypeSequence with all types verifying isDateTime
    // QTypeUnion with at leat one type verifying isDateTime
    public boolean isDateTime() {
        return false;
    }

    // tells whether qtype is a time
    // QTypeAtom with type time
    // QTypeAttribute with type time
    // QTypeElement with type time
    // QTypeCollection with type verifying isTime
    // QTypeSequence with all types verifying isTime
    // QTypeUnion with at leat one type verifying isTime
    public boolean isTime() {
        return false;
    }

    // tells whether qtype is a duration
    // QTypeAtom with type duration
    // QTypeAttribute with type duration
    // QTypeElement with type duration
    // QTypeCollection with type verifying isDuration
    // QTypeSequence with all types verifying isDuration
    // QTypeUnion with at leat one type verifying isDuration
    public boolean isDuration() {
        return false;
    }

    // tells whether qtype is an attribte
    // QTypeAttribute
    // QTypeCollection with type verifying isAttribute
    // QTypeSequence with all types verifying isAttribute
    // QTypeUnion with at leat one type verifying isAttribute
    public boolean isAttribute() {
        return false;
    }

    // tells whether qtype is only an attribte
    // QTypeAttribute
    // QTypeCollection with type verifying isAttribute
    // QTypeSequence with all types verifying isAttribute
    // QTypeUnion with all type verifying isAttribute
    public boolean isOnlyAttribute() {
        return isAttribute();
    }

    // tells whether qtype is a ID, IDREF or IDREFS type
    // QTypeAttribute with special case
    // QTypeCollection with type verifying isIDAttribute
    // QTypeSequence with all types verifying isIDAttribute
    // QTypeUnion with at leat one type verifying isIDAttribute
    public boolean isIDAttribute() {
        return false;
    }

    // tells whether qtype is a node
    // QTypeAttribute
    // QTypeCData
    // QTypeComment
    // QTypeElement
    // QTypeProcessingInstruction
    // QTypeCollection with type verifying isNode
    // QTypeSequence with all types verifying isNode
    // QTypeUnion with at leat one type verifying isNode
    public boolean isNode() {
        return false;
    }

    // tells whether qtype is a node with simple type
    // QTypeAttribute
    // QTypeElement with simple type (typed value)
    // QTypeCollection with type verifying isSimpleTypeNode
    // QTypeSequence with all types verifying isSimpleTypeNode
    // QTypeUnion with at leat one type verifying isSimpleTypeNode
    public boolean isSimpleTypeNode() {
        return false;
    }

    // // tells whether qtype is a collection
    // // QTypeCollection
    // // QTypeUnion with at leat one type verifying isCollection
    // public boolean isCollection() {
    // return false;
    // }

    // tells whether qtype is a collection
    // QTypeCollection
    // QTypeSequence
    // QTypeUnion with at leat one type verifying isMultiple
    public boolean isMultiple() {
        switch (getOccurence()) {
        case OCC_0_0:
        case OCC_1_1:
        case OCC_0_1:
            return false;
        case OCC_0_N:
        case OCC_1_N:
            return true;
        }
        return false;
    }

    // tells whether qtype is atomic
    // QTypeAtom
    // QTypeAttribute
    // QTypeElement with TypedValue not null
    // QTypeUnion with at leat one type verifying isAtomic
    public boolean isAtomic() {
        return false;
    }

    // tells whether qtype is mixed
    // QTypeElement with mixed type (complex type mixed=
    // QTypeCollection with type verifying isMixed
    // QTypeSequence with all types verifying isMixed
    // QTypeUnion with at leat one type verifying isMixed
    public boolean isMixed() {
        return false;
    }

    // tells whether qtype is element
    // QTypeElement
    // QTypeCollection with type verifying isElement
    // QTypeSequence with all types verifying isElement
    // QTypeUnion with at leat one type verifying isElement
    public boolean isElement() {
        return false;
    }

    // tells whether qtype is document
    // QTypeDocument
    // QTypeUnion with at least one type verifying QTypeDocument
    public boolean isDocument() {
        return false;
    }

    // tells whether qtype is qname
    // QTypeAtom with type qname
    public boolean isQName() {
        return false;
    }

    // tells whether the qtype is an empty sequence
    // QTypeSequence with no types
    public boolean isEmptySequence() {
        return false;
    }

    public QType applyDATAFunction() throws TypeException {
        return null;
    }

    public boolean canBeComparedTo(QType compType) {
        return false;
    }

    protected boolean compare(Type type1, Type type2) {
        if (type1 == null || type2 == null)
            return true;
        if (type1 == null && type2 != null)
            return false;
        if (type2 == null && type1 != null)
            return false;
        if (type1 == null || type2 == null || (type1.getName() == null && type2.getName() != null) || (type1.getName() != null && type2.getName() == null))
            return false;
        return ((type1.getName() == null && type2.getName() == null) || (type1.getName() != null && type1.getName().equals(type2.getName()))) && ((type1.getNamespace() == null && type2.getNamespace() == null) || (type1.getNamespace() != null && type1.getNamespace().equals(type2.getNamespace())));
    }

    protected boolean isNumeric(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null) {
            if (type.getName() != null && type.getName().equals(TypeVisitor.SC_ANYTYPE))
                return true;
            return false;
        }
        switch (stype.getPrimitive().getType()) {
        /*
         * case PrimitiveType.STRING : case PrimitiveType.BOOLEAN :
         */
        case PrimitiveType.DECIMAL:
        case PrimitiveType.DOUBLE:
        case PrimitiveType.FLOAT:
            return true;
            /*
             * case PrimitiveType.DURATION : case PrimitiveType.DATE_TIME : case PrimitiveType.TIME : case PrimitiveType.DATE : case PrimitiveType.GYEAR_MONTH : case PrimitiveType.GYEAR : case PrimitiveType.GMONTH_DAY : case PrimitiveType.GDAY : case PrimitiveType.GMONTH : case PrimitiveType.HEX_BINARY : case PrimitiveType.BASE64_BINARY : case PrimitiveType.ANY_URI : case PrimitiveType.QNAME : case PrimitiveType.NOTATION :
             * 
             * case PrimitiveType.ID : case PrimitiveType.IDREF :
             * 
             * case PrimitiveType.LIST : case PrimitiveType.UNION : case PrimitiveType.ANYSIMPLETYPE :
             */
        default:
            break;
        }
        return false;
    }

    protected boolean isInteger(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null)
            return false;
        switch (stype.getPrimitive().getType()) {
        /*
         * case PrimitiveType.STRING : case PrimitiveType.BOOLEAN :
         */
        case PrimitiveType.DECIMAL:
            if (stype.getPrimitive().getFractionDigits() == 0)
                return true;
            /*
             * case PrimitiveType.DOUBLE : case PrimitiveType.FLOAT : return true; case PrimitiveType.DURATION : case PrimitiveType.DATE_TIME : case PrimitiveType.TIME : case PrimitiveType.DATE : case PrimitiveType.GYEAR_MONTH : case PrimitiveType.GYEAR : case PrimitiveType.GMONTH_DAY : case PrimitiveType.GDAY : case PrimitiveType.GMONTH : case PrimitiveType.HEX_BINARY : case PrimitiveType.BASE64_BINARY : case PrimitiveType.ANY_URI : case PrimitiveType.QNAME : case PrimitiveType.NOTATION :
             * 
             * case PrimitiveType.ID : case PrimitiveType.IDREF :
             * 
             * case PrimitiveType.LIST : case PrimitiveType.UNION : case PrimitiveType.ANYSIMPLETYPE :
             */
        default:
            break;
        }
        return false;
    }

    protected boolean isString(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null)
            return false;
        switch (stype.getPrimitive().getType()) {
        case PrimitiveType.STRING:
        case PrimitiveType.ANYSIMPLETYPE:
            return true;
            /*
             * case PrimitiveType.BOOLEAN : case PrimitiveType.DECIMAL : case PrimitiveType.DOUBLE : case PrimitiveType.FLOAT : case PrimitiveType.DURATION : case PrimitiveType.DATE_TIME : case PrimitiveType.TIME : case PrimitiveType.DATE : case PrimitiveType.GYEAR_MONTH : case PrimitiveType.GYEAR : case PrimitiveType.GMONTH_DAY : case PrimitiveType.GDAY : case PrimitiveType.GMONTH : case PrimitiveType.HEX_BINARY : case PrimitiveType.BASE64_BINARY : case PrimitiveType.ANY_URI : case PrimitiveType.QNAME : case PrimitiveType.NOTATION :
             * 
             * case PrimitiveType.ID : case PrimitiveType.IDREF :
             * 
             * case PrimitiveType.LIST : case PrimitiveType.UNION : case PrimitiveType.ANYSIMPLETYPE :
             */
        default:
            break;
        }
        return false;
    }

    protected boolean isDate(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null)
            return false;
        switch (stype.getPrimitive().getType()) {
        // case PrimitiveType.DATE_TIME:
        case PrimitiveType.DATE:
            return true;
        default:
            break;
        }
        return false;
    }

    protected boolean isTime(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null)
            return false;
        switch (stype.getPrimitive().getType()) {
        // case PrimitiveType.DATE_TIME:
        case PrimitiveType.TIME:
            return true;
        default:
            break;
        }
        return false;
    }

    protected boolean isDateTime(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null)
            return false;
        switch (stype.getPrimitive().getType()) {
        case PrimitiveType.DATE:
        case PrimitiveType.DATE_TIME:
        case PrimitiveType.TIME:
            return true;
        default:
            break;
        }
        return false;
    }

    protected boolean isDuration(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null)
            return false;
        switch (stype.getPrimitive().getType()) {
        case PrimitiveType.DURATION:
             return true;
        default:
            break;
        }
        return false;
    }

    protected boolean isBoolean(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null)
            return false;
        switch (stype.getPrimitive().getType()) {
        // case PrimitiveType.STRING :
        case PrimitiveType.BOOLEAN:
            return true;
            /*
             * case PrimitiveType.DECIMAL : case PrimitiveType.DOUBLE : case PrimitiveType.FLOAT : case PrimitiveType.DURATION : case PrimitiveType.DATE_TIME : case PrimitiveType.TIME : case PrimitiveType.DATE : case PrimitiveType.GYEAR_MONTH : case PrimitiveType.GYEAR : case PrimitiveType.GMONTH_DAY : case PrimitiveType.GDAY : case PrimitiveType.GMONTH : case PrimitiveType.HEX_BINARY : case PrimitiveType.BASE64_BINARY : case PrimitiveType.ANY_URI : case PrimitiveType.QNAME : case PrimitiveType.NOTATION :
             * 
             * case PrimitiveType.ID : case PrimitiveType.IDREF :
             * 
             * case PrimitiveType.LIST : case PrimitiveType.UNION : case PrimitiveType.ANYSIMPLETYPE :
             */
        default:
            break;
        }
        return false;
    }

    protected boolean isQName(Type type) {
        SimpleType stype = type.getValueType();
        if (stype == null)
            return false;
        switch (stype.getPrimitive().getType()) {
        case PrimitiveType.STRING:
            /*
             * case PrimitiveType.BOOLEAN : case PrimitiveType.DECIMAL : case PrimitiveType.DOUBLE : case PrimitiveType.FLOAT : case PrimitiveType.DURATION : case PrimitiveType.DATE_TIME : case PrimitiveType.TIME : case PrimitiveType.DATE : case PrimitiveType.GYEAR_MONTH : case PrimitiveType.GYEAR : case PrimitiveType.GMONTH_DAY : case PrimitiveType.GDAY : case PrimitiveType.GMONTH : case PrimitiveType.HEX_BINARY : case PrimitiveType.BASE64_BINARY : case PrimitiveType.ANY_URI :
             */
        case PrimitiveType.QNAME:
            return true;
            /*
             * case PrimitiveType.NOTATION :
             * 
             * case PrimitiveType.ID : case PrimitiveType.IDREF :
             * 
             * case PrimitiveType.LIST : case PrimitiveType.UNION : case PrimitiveType.ANYSIMPLETYPE :
             */
        default:
            break;
        }
        return false;
    }

    protected boolean canBeCompared(PrimitiveType type1, PrimitiveType type2) {
        int primitiveType1 = type1.getType();
        int primitiveType2 = type2.getType();
        if (primitiveType1 == PrimitiveType.ANYSIMPLETYPE || primitiveType2 == PrimitiveType.ANYSIMPLETYPE)
            return true;
        if (primitiveType1 == primitiveType2)
            return true;
        switch (primitiveType1) {
        case PrimitiveType.ANY_URI :
        case PrimitiveType.STRING :
            if (primitiveType2 == PrimitiveType.ANY_URI)
                return true;
            if (primitiveType2 == PrimitiveType.STRING)
                return true;
            break;
        // case PrimitiveType.BOOLEAN :
        case PrimitiveType.DECIMAL:
        case PrimitiveType.DOUBLE:
        case PrimitiveType.FLOAT:
            if (primitiveType2 == PrimitiveType.DECIMAL)
                return true;
            if (primitiveType2 == PrimitiveType.DOUBLE)
                return true;
            if (primitiveType2 == PrimitiveType.FLOAT)
                return true;
            break;
        // case PrimitiveType.DURATION :
        case PrimitiveType.DATE_TIME:
        case PrimitiveType.TIME:
        case PrimitiveType.DATE:
        case PrimitiveType.GYEAR_MONTH:
        case PrimitiveType.GYEAR:
        case PrimitiveType.GMONTH_DAY:
        case PrimitiveType.GDAY:
        case PrimitiveType.GMONTH:
            if (primitiveType2 == PrimitiveType.DATE_TIME)
                return true;
            if (primitiveType2 == PrimitiveType.TIME)
                return true;
            if (primitiveType2 == PrimitiveType.DATE)
                return true;
            if (primitiveType2 == PrimitiveType.GYEAR_MONTH)
                return true;
            if (primitiveType2 == PrimitiveType.GYEAR)
                return true;
            if (primitiveType2 == PrimitiveType.GMONTH_DAY)
                return true;
            if (primitiveType2 == PrimitiveType.GDAY)
                return true;
            if (primitiveType2 == PrimitiveType.GMONTH)
                return true;
            break;
        // case PrimitiveType.HEX_BINARY :
        // case PrimitiveType.BASE64_BINARY :
        // case PrimitiveType.QNAME :
        // case PrimitiveType.NOTATION :
        // case PrimitiveType.ID :
        // case PrimitiveType.IDREF :
        // case PrimitiveType.LIST :
        case PrimitiveType.UNION :
            List primitives = ((PrimitiveUnionType)type1).getPrimitives();
            for (int i=0;i<primitives.size();i++) {
                if (canBeCompared(type2,(PrimitiveType)primitives.get(i)))
                    return true;
            }
            break;
        // case PrimitiveType.ANYSIMPLETYPE :
        default:
            break;
        }
        return false;
    }

    abstract boolean canBeCastedInto(QType castType, byte occurence);

    public boolean canBeCastedInto(QType castType) {
        return canBeCastedInto(castType, castType.getOccurence());
    }

    public boolean canBeCastedInto(int castType) {
        return false;
    }

    static public boolean canBeCastedInto(int castType1, int castType2) {
        if (castType1 == -1 || castType2 == -1)
            return false;
        return castTab[castType1][castType2];
    }

    static public boolean canBeCastedInto(byte occurence1, byte occurence2) {
        if (occurence1 == occurence2)
            return true;
        switch (occurence1) {
        case OCC_0_0:
            if (occurence2 != OCC_1_1 && occurence2 != OCC_1_N)
                return true;
            break;
        case OCC_0_1:
            if (occurence2 == OCC_0_N)
                return true;
            break;
        case OCC_1_1:
            if (occurence2 != OCC_0_0)
                return true;
            break;
        }
        return false;
    }

    static public byte computeComplexOccurence(byte occurence1, byte occurence2) {
        if (occurence1 == OCC_0_0 || occurence2 == OCC_0_0)
            return OCC_0_0;
        if (occurence1 == OCC_1_1)
            return occurence2;
        if (occurence2 == OCC_1_1)
            return occurence1;
        if (occurence1 == OCC_0_1 && occurence2 == OCC_0_1)
            return OCC_0_1;
        if (occurence1 == OCC_1_N && occurence2 == OCC_1_N)
            return OCC_1_N;
        return OCC_0_N;
    }

    // public int getCastType() {
    // return -1;
    // }
    static public int getCastType(Type type) {
        if (type == null)
            return QType.PRIMITIVE_anySimpleType;
        if (!(type instanceof SimpleType))
            return -1;
        switch (((SimpleType) type).getPrimitive().getType()) {
        // case PrimitiveType.VOID : return -1;
        case PrimitiveType.ANYSIMPLETYPE:
            return QType.PRIMITIVE_anySimpleType;
        case PrimitiveType.STRING:
            return QType.PRIMITIVE_string;
        case PrimitiveType.BOOLEAN:
            return QType.PRIMITIVE_boolean;
        case PrimitiveType.DECIMAL:
            return QType.PRIMITIVE_decimal;
        case PrimitiveType.DOUBLE:
            return QType.PRIMITIVE_double;
        case PrimitiveType.FLOAT:
            return QType.PRIMITIVE_float;
        case PrimitiveType.DURATION:
            return QType.PRIMITIVE_duration;
        case PrimitiveType.DATE_TIME:
            return QType.PRIMITIVE_dateTime;
        case PrimitiveType.TIME:
            return QType.PRIMITIVE_time;
        case PrimitiveType.DATE:
            return QType.PRIMITIVE_date;
        case PrimitiveType.GYEAR_MONTH:
            return QType.PRIMITIVE_gYearMonth;
        case PrimitiveType.GYEAR:
            return QType.PRIMITIVE_gYear;
        case PrimitiveType.GMONTH_DAY:
            return QType.PRIMITIVE_gMonthDay;
        case PrimitiveType.GDAY:
            return QType.PRIMITIVE_gDay;
        case PrimitiveType.GMONTH:
            return QType.PRIMITIVE_gMonth;
        case PrimitiveType.HEX_BINARY:
            return QType.PRIMITIVE_hexBinary;
        case PrimitiveType.BASE64_BINARY:
            return QType.PRIMITIVE_base64Binary;
        case PrimitiveType.ANY_URI:
            return QType.PRIMITIVE_anyURI;
        case PrimitiveType.QNAME:
            return QType.PRIMITIVE_Qname;
        case PrimitiveType.NOTATION:
            return QType.PRIMITIVE_NOTATION;
        }
        return -1;
    }

    /*
     * // not bounded values !!! public static final int PRIMITIVE_anySimpleType = 1; public static final int PRIMITIVE_dayTimeDuration = 10; public static final int PRIMITIVE_yearMonthDuration = 21;
     */

    // methods added to implement xqj interface
    // Returns true if the given type exactly matches or is a supertype of this type.
    public boolean isOfType(QType type) {
        if (exactlyMatches(type))
            return true;
        if (type.isSuperTypeOf(this))
            return true;
        return false;
    }

    public boolean exactlyMatches(QType type) {
        if (type == null)
            return false;
        if (this.subclass != type.subclass)
            return false;
        if (this.occurence != type.occurence)
            return false;
        // if (!this.getType().equals(type.getType())) return false;
        return true;
    }

    public boolean isSuperTypeOf(QType type) {
        if (type.canBeCastedInto(this))
            return true;
        // TODO further explore super type relation
        if (this.subclass != type.subclass)
            return false;
        switch (type.occurence) {
        case OCC_0_0:
            if (occurence == OCC_0_0 || occurence == OCC_0_1 || occurence == OCC_0_N)
                return true;
            break;
        case OCC_0_1:
            if (occurence == OCC_0_1 || occurence == OCC_0_N)
                return true;
            break;
        case OCC_0_N:
            if (occurence == OCC_0_N)
                return true;
            break;
        case OCC_1_1:
            if (occurence != OCC_0_0)
                return true;
            break;
        case OCC_1_N:
            if (occurence == OCC_1_N || occurence == OCC_0_N)
                return true;
            break;
        }
        return false;
    }

    public boolean isTestType(byte testType) {
        switch (testType) {
        // case PrimitiveType.VOID : return -1;
        case QType.PRIMITIVE_anySimpleType:
            return false;
        case QType.PRIMITIVE_string:
            return isString();
        case QType.PRIMITIVE_boolean:
            return isBoolean();
        case QType.PRIMITIVE_decimal:
            return false;
        case QType.PRIMITIVE_double:
            return false;
        case QType.PRIMITIVE_float:
            return false;
        case QType.PRIMITIVE_dayTimeDuration:
            return isDuration();
        case QType.PRIMITIVE_dateTime:
            return isDateTime();
        case QType.PRIMITIVE_time:
            return isTime();
        case QType.PRIMITIVE_date:
            return isDate();
        case QType.PRIMITIVE_gYearMonth:
            return false;
        case QType.PRIMITIVE_gYear:
            return false;
        case QType.PRIMITIVE_gMonthDay:
            return false;
        case QType.PRIMITIVE_gDay:
            return false;
        case QType.PRIMITIVE_gMonth:
            return false;
        case QType.PRIMITIVE_hexBinary:
            return false;
        case QType.PRIMITIVE_base64Binary:
            return false;
        case QType.PRIMITIVE_anyURI:
            return false;
        case QType.PRIMITIVE_Qname:
            return false;
        case QType.PRIMITIVE_NOTATION:
            return false;
        }
        return false;
    }
}
