/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;

import org.xquark.xquery.parser.XQueryException;

public class QTypeSequence extends QTypePrime implements QTypeVisitable {
	private static final String RCSRevision = "$Revision: 1.7 $";
	private static final String RCSName = "$Name:  $";

	List list = null;

	public QTypeSequence(List list) {
		super(list,SEQUENCE);
		setList(list);
		this.occurence = OCC_1_1;
	}
	public QTypeSequence(List list, byte occurence) {
		super(list,SEQUENCE);
		setList(list);
		this.occurence = occurence;
	}

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

    public void accept(QTypeVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// get/set
	public List getList() {
		return list;
	}

	private void setList(List list) {
		//if (list == null) throw new TypeException("list of QTypeSequence cannot be null");
		this.list = list;
	}

	// clone
	public Object clone() throws CloneNotSupportedException {
		QTypeSequence newObj = new QTypeSequence(list,occurence);
		return newObj;
	}
	// print
	public String toSimpleString() {
		return "QTypeSequence(" + list + ")";
	}

	public String toString() {
		return "QTypeSequence(" + list + ")";
	}
    
    // methods added to implement xqj interface
    public int getQTypeOccurence(QType qtype) {
        if (qtype == null) return 0;
        int retValue = 0;
        for (Iterator it = this.list.iterator();it.hasNext();) {
            if (qtype.exactlyMatches((QType)it.next())) retValue++;
        }
        return retValue;
    }

	public boolean canBeCastedInto(QType castType, byte occurence) {
		if (castType == null)
			return false;
		if (castType instanceof QTypeUnion) {
			QTypeUnion qtypeUnion = (QTypeUnion) castType;
			for (int i = 0; i < qtypeUnion.getList().size(); i++) {
				QType tmpQType = (QType) ((QTypeUnion) castType).getList().get(i);
				byte newOccurence = computeComplexOccurence(occurence, tmpQType.getOccurence());
				if (canBeCastedInto(this.occurence, newOccurence))
					if (canBeCastedInto(tmpQType, newOccurence))
						return true;
			}
		}
		if (!(castType instanceof QTypeSequence))
			return false;
		if (!canBeCastedInto(this.occurence, occurence))
			return false;
		QTypeSequence qtypeSequence = (QTypeSequence)castType;
		if (this.getList().size() != qtypeSequence.getList().size())
			return false;
		for (int i=0;i<this.getList().size();i++) {
			if (!((QType)this.getList().get(i)).canBeCastedInto((QType)qtypeSequence.getList().get(i)))
				return false;
		}
		return true;
	}

    // tells whether the qtype is an empty sequence
    // QTypeSequence with no types
    public boolean isEmptySequence() {
         return (list == null || list.isEmpty());
    }

}
