/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.xquark.schema.SimpleType;
import org.xquark.xquery.parser.XQueryException;

public class QTypeUnion extends QTypePrime implements QTypeVisitable {
	private static final String RCSRevision = "$Revision: 1.6 $";

	private static final String RCSName = "$Name:  $";

	List list = null;

	public QTypeUnion(List list) {
		super(list, UNION);
		setList(list);
		subclass = UNION;
		// primeList = list;
		occurence = OCC_1_1;
	}

	public QTypeUnion(List list, byte occurence) {
		super(list, UNION);
		setList(list);
		subclass = UNION;
		// primeList = list;
		this.occurence = occurence;
	}

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(QTypeVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// get/set
	public List getList() {
		return list;
	}

	private void setList(List list) {
		// if (list == null || list.isEmpty()) throw new TypeException("list of QTypeUnion cannot be null or empty");
		this.list = new ArrayList();
		Iterator it = list.iterator();
		while (it.hasNext()) {
			QType qtype = (QType) it.next();
			if (qtype instanceof QTypeUnion)
				this.list.addAll(((QTypeUnion) qtype).getList());
			else
				this.list.add(qtype);
		}
	}

	// clone
	public Object clone() throws CloneNotSupportedException {
		QTypeUnion newObj = new QTypeUnion(list, occurence);
		return newObj;
	}

	// print
	public String toSimpleString() {
		return "QTypeUnion(" + list + ")";
	}

	public String toString() {
		return "QTypeUnion(" + list + ")";
	}

	// tells whether qtype is element
	public SimpleType getSimpleType() {
		Iterator it = primeList.iterator();
		SimpleType retType = null;
		while (it.hasNext()) {
			QType qtype = (QType) it.next();
			SimpleType stype = qtype.getSimpleType();
			if (stype != null) {
				if (retType != null) {
					if (retType != stype)
						return (SimpleType) stype.getSchema().getType("AnySimpleType");
				} else
					retType = stype;
			}
		}
		return retType;
	}

	public boolean canBeCastedInto(QType castType, byte occurence) {
		if (castType == null)
			return false;
		if (castType instanceof QTypeSequence) {
			if (((QTypeSequence) castType).getList().size() == 1) {
				QType tmpQType = (QType) ((QTypeSequence) castType).getList().get(0);
				byte newOccurence = computeComplexOccurence(occurence, tmpQType.getOccurence());
				if (canBeCastedInto(this.occurence, newOccurence))
					return canBeCastedInto(tmpQType, newOccurence);
			}
		}
//		if (!(castType instanceof QTypeUnion))
//			return false;
		if (!canBeCastedInto(this.occurence, occurence))
			return false;
		for (int i = 0; i < this.getPrimeList().size(); i++) {
			QType tmpQType = (QType) this.getPrimeList().get(i);
			if (!tmpQType.canBeCastedInto(castType, QType.OCC_1_1))
				return false;
		}
		return true;
	}

}
